# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import pathlib
from typing import List

from generator import model

from .rust_commons import TypeData, generate_commons
from .rust_enum import generate_enums
from .rust_file_header import license_header
from .rust_lang_utils import lines_to_comments
from .rust_structs import (
    generate_notifications,
    generate_requests,
    generate_structures,
    generate_type_aliases,
)

PACKAGE_DIR_NAME = "lsprotocol"


def generate_from_spec(spec: model.LSPModel, output_dir: str) -> None:
    code = generate_package_code(spec)

    output_path = pathlib.Path(output_dir, PACKAGE_DIR_NAME)
    if not output_path.exists():
        output_path.mkdir(parents=True, exist_ok=True)
        (output_path / "src").mkdir(parents=True, exist_ok=True)

    for file_name in code:
        (output_path / file_name).write_text(code[file_name], encoding="utf-8")


def generate_package_code(spec: model.LSPModel) -> List[str]:
    return {
        "src/lib.rs": generate_lib_rs(spec),
    }


def generate_lib_rs(spec: model.LSPModel) -> List[str]:
    lines = lines_to_comments(license_header())
    lines += [
        "",
        "// ****** THIS IS A GENERATED FILE, DO NOT EDIT. ******",
        "// Steps to generate:",
        "// 1. Checkout https://github.com/microsoft/lsprotocol",
        "// 2. Install nox: `python -m pip install nox`",
        "// 3. Run command: `python -m nox --session build_lsp`",
        "",
    ]
    lines += [
        "use serde::{Serialize, Deserialize};",
        "use std::collections::HashMap;",
        "use rust_decimal::Decimal;" "",
    ]

    type_data = TypeData()
    generate_commons(spec, type_data)
    generate_enums(spec.enumerations, type_data)

    generate_type_aliases(spec, type_data)
    generate_structures(spec, type_data)
    generate_notifications(spec, type_data)
    generate_requests(spec, type_data)

    lines += type_data.get_lines()
    return "\n".join(lines)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import hamcrest
import pytest

from lsprotocol import types as lsp


@pytest.mark.parametrize(
    ("a", "b", "expected"),
    [
        (
            lsp.Location(
                "some_path", lsp.Range(lsp.Position(1, 23), lsp.Position(4, 56))
            ),
            lsp.Location(
                "some_path", lsp.Range(lsp.Position(1, 23), lsp.Position(4, 56))
            ),
            True,
        ),
        (
            lsp.Location(
                "some_path", lsp.Range(lsp.Position(1, 23), lsp.Position(4, 56))
            ),
            lsp.Location(
                "some_path2", lsp.Range(lsp.Position(1, 23), lsp.Position(4, 56))
            ),
            False,
        ),
        (
            lsp.Location(
                "some_path", lsp.Range(lsp.Position(1, 23), lsp.Position(4, 56))
            ),
            lsp.Location(
                "some_path", lsp.Range(lsp.Position(1, 23), lsp.Position(8, 91))
            ),
            False,
        ),
    ],
)
def test_location_equality(a, b, expected):
    hamcrest.assert_that(a == b, hamcrest.is_(expected))


def test_location_repr():
    a = lsp.Location("some_path", lsp.Range(lsp.Position(1, 23), lsp.Position(4, 56)))
    hamcrest.assert_that(f"{a!r}", hamcrest.is_("some_path:1:23-4:56"))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json
import pathlib
from typing import Generator, List, Union

import pytest

import lsprotocol.converters as cv
import lsprotocol.types as lsp

TEST_DATA_ROOT = pathlib.Path(__file__).parent.parent.parent / "packages" / "testdata"


def get_all_json_files(root: Union[pathlib.Path, str]) -> List[pathlib.Path]:
    root_path = pathlib.Path(root)
    return list(root_path.glob("**/*.json"))


converter = cv.get_converter()


@pytest.mark.parametrize("json_file", get_all_json_files(TEST_DATA_ROOT))
def test_generated_data(json_file: str) -> None:
    type_name, result_type, _ = json_file.name.split("-", 2)
    lsp_type = getattr(lsp, type_name)
    data = json.loads(json_file.read_text(encoding="utf-8"))

    try:
        converter.structure(data, lsp_type)
        assert result_type == "True", "Expected error, but succeeded structuring"
    except Exception as e:
        assert result_type == "False", "Expected success, but failed structuring"



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json
import uuid

import hamcrest
import jsonrpc
import pytest
from cattrs.errors import ClassValidationError

ID = str(uuid.uuid4())

TEST_DATA = [
    (
        {"id": ID, "method": "workspace/semanticTokens/refresh", "jsonrpc": "2.0"},
        json.dumps(
            {"id": ID, "method": "workspace/semanticTokens/refresh", "jsonrpc": "2.0"}
        ),
    ),
    (
        {
            "id": ID,
            "method": "workspace/semanticTokens/refresh",
            "params": None,
            "jsonrpc": "2.0",
        },
        json.dumps(
            {"id": ID, "method": "workspace/semanticTokens/refresh", "jsonrpc": "2.0"}
        ),
    ),
]


@pytest.mark.parametrize("index", list(range(0, len(TEST_DATA))))
def test_workspace_sematic_tokens_refresh_request_serialization(index):
    data, expected = TEST_DATA[index]
    data_str = json.dumps(data)
    parsed = jsonrpc.from_json(data_str)
    actual_str = jsonrpc.to_json(parsed)
    hamcrest.assert_that(actual_str, hamcrest.is_(expected))


@pytest.mark.parametrize(
    "data",
    [
        json.dumps({}),  # missing method and jsonrpc
        json.dumps({"method": "invalid"}),  # invalid method type
    ],
)
def test_workspace_sematic_tokens_refresh_request_invalid(data):
    with pytest.raises((ClassValidationError, KeyError)):
        jsonrpc.from_json(data)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Dict, List, Tuple, Union

from generator import model

TypesWithId = Union[
    model.Request,
    model.TypeAlias,
    model.Enum,
    model.Structure,
    model.Notification,
    model.LiteralType,
    model.ReferenceType,
    model.ReferenceMapKeyType,
    model.Property,
    model.EnumItem,
]


class TypeData:
    def __init__(self) -> None:
        self._id_data: Dict[str, Tuple[str, TypesWithId, List[str]]] = {}
        self._ctor_data: Dict[str, Tuple[str, str]] = {}

    def add_type_info(
        self,
        type_def: TypesWithId,
        type_name: str,
        impl: List[str],
    ) -> None:
        if type_def.id_ in self._id_data:
            raise Exception(f"Duplicate id {type_def.id_} for type {type_name}")
        self._id_data[type_def.id_] = (type_name, type_def, impl)

    def has_id(
        self,
        type_def: TypesWithId,
    ) -> bool:
        return type_def.id_ in self._id_data

    def has_name(self, type_name: str) -> bool:
        return any(type_name == name for name, _, _ in self._id_data.values())

    def get_by_name(self, type_name: str) -> List[TypesWithId]:
        return [
            type_def
            for name, type_def, _ in self._id_data.values()
            if name == type_name
        ]

    def get_all(self) -> List[Tuple[str, List[str]]]:
        return [(name, lines) for name, _, lines in self._id_data.values()]

    def add_ctor(self, type_name: str, ctor: Tuple[str, str, bool]) -> None:
        self._ctor_data[type_name] = ctor

    def get_ctor(self, type_name: str) -> Tuple[str, str, bool]:
        return self._ctor_data[type_name]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json

import hamcrest
import jsonrpc
import pytest

from lsprotocol.types import (
    ProgressNotification,
    ProgressParams,
    WorkDoneProgressBegin,
    WorkDoneProgressEnd,
    WorkDoneProgressReport,
)


@pytest.mark.parametrize(
    "obj, expected",
    [
        (
            ProgressNotification(
                params=ProgressParams(
                    token="id1",
                    value=WorkDoneProgressBegin(title="Begin Progress", percentage=0),
                )
            ),
            json.dumps(
                {
                    "params": {
                        "token": "id1",
                        "value": {
                            "title": "Begin Progress",
                            "kind": "begin",
                            "percentage": 0,
                        },
                    },
                    "method": "$/progress",
                    "jsonrpc": "2.0",
                }
            ),
        ),
        (
            ProgressNotification(
                params=ProgressParams(
                    token="id1",
                    value=WorkDoneProgressReport(message="Still going", percentage=50),
                )
            ),
            json.dumps(
                {
                    "params": {
                        "token": "id1",
                        "value": {
                            "kind": "report",
                            "message": "Still going",
                            "percentage": 50,
                        },
                    },
                    "method": "$/progress",
                    "jsonrpc": "2.0",
                }
            ),
        ),
        (
            ProgressNotification(
                params=ProgressParams(
                    token="id1",
                    value=WorkDoneProgressEnd(message="Finished"),
                )
            ),
            json.dumps(
                {
                    "params": {
                        "token": "id1",
                        "value": {
                            "kind": "end",
                            "message": "Finished",
                        },
                    },
                    "method": "$/progress",
                    "jsonrpc": "2.0",
                }
            ),
        ),
    ],
)
def test_exit_serialization(obj, expected):
    actual_str = jsonrpc.to_json(obj)
    hamcrest.assert_that(actual_str, hamcrest.is_(expected))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json

import hamcrest
import jsonrpc
import pytest
from cattrs.errors import ClassValidationError


@pytest.mark.parametrize(
    "data, expected",
    [
        (
            {"method": "exit", "jsonrpc": "2.0"},
            json.dumps({"method": "exit", "jsonrpc": "2.0"}),
        ),
        (
            {"method": "exit", "params": None, "jsonrpc": "2.0"},
            json.dumps({"method": "exit", "jsonrpc": "2.0"}),
        ),
    ],
)
def test_exit_serialization(data, expected):
    data_str = json.dumps(data)
    parsed = jsonrpc.from_json(data_str)
    actual_str = jsonrpc.to_json(parsed)
    hamcrest.assert_that(actual_str, hamcrest.is_(expected))


@pytest.mark.parametrize(
    "data",
    [
        json.dumps({}),  # missing method and jsonrpc
        json.dumps({"method": "invalid"}),  # invalid method type
    ],
)
def test_exit_invalid(data):
    with pytest.raises((ClassValidationError, KeyError)):
        jsonrpc.from_json(data)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import List, Union

import generator.model as model

from .rust_commons import TypeData, generate_extras
from .rust_lang_utils import indent_lines, lines_to_doc_comments, to_upper_camel_case


def _get_enum_docs(enum: Union[model.Enum, model.EnumItem]) -> List[str]:
    doc = enum.documentation.splitlines(keepends=False) if enum.documentation else []
    return lines_to_doc_comments(doc)


def generate_enum(enum: model.Enum, types: TypeData) -> None:
    is_int = all(isinstance(item.value, int) for item in enum.values)

    lines = (
        _get_enum_docs(enum)
        + generate_extras(enum)
        + [
            "#[derive(Serialize, Deserialize, PartialEq, Debug, Eq, Clone)]",
            f"pub enum {enum.name} " "{",
        ]
    )

    for item in enum.values:
        if is_int:
            field = [
                f"{to_upper_camel_case(item.name)} = {item.value},",
            ]
        else:
            field = [
                f'#[serde(rename = "{item.value}")]',
                f"{to_upper_camel_case(item.name)},",
            ]

        lines += indent_lines(
            _get_enum_docs(item) + generate_extras(item) + field + [""]
        )

    lines += ["}"]

    types.add_type_info(enum, enum.name, lines)


def generate_enums(enums: List[model.Enum], types: TypeData) -> None:
    for enum in enums:
        generate_enum(enum, types)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import hamcrest
import pytest

from lsprotocol import types as lsp


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("refactor", lsp.CodeActionKind.Refactor),
        (lsp.CodeActionKind.Refactor, lsp.CodeActionKind.Refactor),
        ("namespace", lsp.SemanticTokenTypes.Namespace),
        (lsp.SemanticTokenTypes.Namespace, lsp.SemanticTokenTypes.Namespace),
        ("declaration", lsp.SemanticTokenModifiers.Declaration),
        (
            lsp.SemanticTokenModifiers.Declaration,
            lsp.SemanticTokenModifiers.Declaration,
        ),
        ("comment", lsp.FoldingRangeKind.Comment),
        (lsp.FoldingRangeKind.Comment, lsp.FoldingRangeKind.Comment),
        ("utf-8", lsp.PositionEncodingKind.Utf8),
        (lsp.PositionEncodingKind.Utf8, lsp.PositionEncodingKind.Utf8),
        (1, lsp.WatchKind.Create),
        (lsp.WatchKind.Create, lsp.WatchKind.Create),
        (-32700, lsp.ErrorCodes.ParseError),
        (lsp.ErrorCodes.ParseError, lsp.ErrorCodes.ParseError),
        (-32803, lsp.LSPErrorCodes.RequestFailed),
        (lsp.LSPErrorCodes.RequestFailed, lsp.LSPErrorCodes.RequestFailed),
    ],
)
def test_custom_enum_types(value, expected):
    hamcrest.assert_that(value, hamcrest.is_(expected))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json

from lsprotocol import converters, types


def to_json(
    obj: types.MESSAGE_TYPES,
    method: str = None,
    converter=None,
) -> str:
    """Converts a given LSP message object to JSON string using the provided
    converter."""
    if not converter:
        converter = converters.get_converter()

    if method is None:
        method = obj.method if hasattr(obj, "method") else None

    if hasattr(obj, "result"):
        if method is None:
            raise ValueError(f"`method` must not be None for response type objects.")
        obj_type = types.METHOD_TO_TYPES[method][1]
    elif hasattr(obj, "error"):
        obj_type = types.ResponseErrorMessage
    else:
        obj_type = types.METHOD_TO_TYPES[method][0]
    return json.dumps(converter.unstructure(obj, unstructure_as=obj_type))


def from_json(json_str: str, method: str = None, converter=None) -> types.MESSAGE_TYPES:
    """Parses and given JSON string and returns LSP message object using the provided
    converter."""
    if not converter:
        converter = converters.get_converter()

    obj = json.loads(json_str)

    if method is None:
        method = obj.get("method", None)

    if "result" in obj:
        if method is None:
            raise ValueError(f"`method` must not be None for response type objects.")
        obj_type = types.METHOD_TO_TYPES[method][1]
    elif "error" in obj:
        obj_type = types.ResponseErrorMessage
    else:
        obj_type = types.METHOD_TO_TYPES[method][0]
    return converter.structure(obj, obj_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import hamcrest
import pytest
from cattrs import ClassValidationError

from lsprotocol import converters as cv
from lsprotocol import types as lsp

TEST_DATA = [
    {
        "id": 1,
        "result": [{"name": "test", "kind": 1, "location": {"uri": "test"}}],
        "jsonrpc": "2.0",
    },
    {
        "id": 1,
        "result": [
            {
                "name": "test",
                "kind": 1,
                "location": {
                    "uri": "test",
                    "range": {
                        "start": {"line": 1, "character": 1},
                        "end": {"line": 1, "character": 1},
                    },
                },
            }
        ],
        "jsonrpc": "2.0",
    },
    {
        "id": 1,
        "result": [{"name": "test", "kind": 1, "location": {"uri": "test"}, "data": 1}],
        "jsonrpc": "2.0",
    },
    {
        "id": 1,
        "result": [
            {
                "name": "test",
                "kind": 1,
                "location": {
                    "uri": "test",
                    "range": {
                        "start": {"line": 1, "character": 1},
                        "end": {"line": 1, "character": 1},
                    },
                },
                "deprecated": True,
            }
        ],
        "jsonrpc": "2.0",
    },
]

BAD_TEST_DATA = [
    {
        "id": 1,
        "result": [
            {
                "name": "test",
                "kind": 1,
                "location": {"uri": "test"},
                "deprecated": True,
            }
        ],
        "jsonrpc": "2.0",
    },
]


@pytest.mark.parametrize("data", TEST_DATA)
def test_workspace_symbols(data):
    converter = cv.get_converter()
    obj = converter.structure(data, lsp.WorkspaceSymbolResponse)
    hamcrest.assert_that(obj, hamcrest.instance_of(lsp.WorkspaceSymbolResponse))
    hamcrest.assert_that(
        converter.unstructure(obj, lsp.WorkspaceSymbolResponse),
        hamcrest.is_(data),
    )


@pytest.mark.parametrize("data", BAD_TEST_DATA)
def test_workspace_symbols_bad(data):
    converter = cv.get_converter()
    with pytest.raises(ClassValidationError):
        obj = converter.structure(data, lsp.WorkspaceSymbolResponse)
        hamcrest.assert_that(obj, hamcrest.instance_of(lsp.WorkspaceSymbolResponse))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import re
from typing import List, Optional, Union

from generator import model

BASIC_LINK_RE = re.compile(r"{@link +(\w+) ([\w ]+)}")
BASIC_LINK_RE2 = re.compile(r"{@link +(\w+)\.(\w+) ([\w \.`]+)}")
BASIC_LINK_RE3 = re.compile(r"{@link +(\w+)}")
BASIC_LINK_RE4 = re.compile(r"{@link +(\w+)\.(\w+)}")
PARTS_RE = re.compile(r"(([a-z0-9])([A-Z]))")


def _fix_links(line: str) -> str:
    line = BASIC_LINK_RE.sub(r'<see cref="\1">\2</see>', line)
    line = BASIC_LINK_RE2.sub(r'<see cref="\1.\2">\3</see>', line)
    line = BASIC_LINK_RE3.sub(r'<see cref="\1" />', line)
    line = BASIC_LINK_RE4.sub(r'<see cref="\1.\2" />', line)
    return line


def lines_to_doc_comments(lines: List[str]) -> List[str]:
    if not lines:
        return []

    return (
        ["/// <summary>"]
        + [f"/// {_fix_links(line)}" for line in lines if not line.startswith("@")]
        + ["/// </summary>"]
    )


def get_parts(name: str) -> List[str]:
    name = name.replace("_", " ")
    return PARTS_RE.sub(r"\2 \3", name).split()


def to_camel_case(name: str) -> str:
    parts = get_parts(name)
    return parts[0] + "".join([p.capitalize() for p in parts[1:]])


def to_upper_camel_case(name: str) -> str:
    return "".join([c.capitalize() for c in get_parts(name)])


def lsp_method_to_name(method: str) -> str:
    if method.startswith("$"):
        method = method[1:]
    method = method.replace("/", "_")
    return to_upper_camel_case(method)


def file_header() -> List[str]:
    return [
        "// Copyright (c) Microsoft Corporation. All rights reserved.",
        "// Licensed under the MIT License.",
        "// ",
        "// THIS FILE IS AUTOGENERATED, DO NOT MODIFY IT",
        "",
    ]


def namespace_wrapper(
    namespace: str, imports: List[str], lines: List[str]
) -> List[str]:
    indent = " " * 4
    return (
        file_header()
        + imports
        + [""]
        + ["namespace " + namespace + " {"]
        + [(f"{indent}{line}" if line else line) for line in lines]
        + ["}", ""]
    )


def get_doc(doc: Optional[str]) -> str:
    if doc:
        return lines_to_doc_comments(doc.splitlines(keepends=False))
    return []


def get_special_case_class_name(name: str) -> str:
    # This is because C# does not allow class name and property name to be the same.
    # public class Command{ public string Command { get; set; }} is not valid.
    if name == "Command":
        return "CommandAction"
    return name


def get_special_case_property_name(name: str) -> str:
    if name == "string":
        return "stringValue"
    if name == "int":
        return "intValue"
    if name == "event":
        return "eventArgs"
    if name == "params":
        return "paramsValue"
    return name


def class_wrapper(
    type_def: Union[model.Structure, model.Notification, model.Request],
    inner: List[str],
    derived: Optional[str] = None,
    class_attributes: Optional[List[str]] = None,
    is_record=True,
) -> List[str]:
    if hasattr(type_def, "name"):
        name = get_special_case_class_name(type_def.name)
    else:
        raise ValueError(f"Unknown type: {type_def}")

    rec_or_cls = "record" if is_record else "class"
    lines = (
        get_doc(type_def.documentation)
        + generate_extras(type_def)
        + (class_attributes if class_attributes else [])
        + [
            "[DataContract]",
            f"public {rec_or_cls} {name}: {derived}"
            if derived
            else f"public {rec_or_cls} {name}",
            "{",
        ]
    )
    lines += indent_lines(inner)
    lines += ["}", ""]
    return lines


def property_wrapper(prop_def: model.Property, content: List[str]) -> List[str]:
    lines = (get_doc(prop_def.documentation) + generate_extras(prop_def) + content,)
    lines += indent_lines(content)
    return lines


def indent_lines(lines: List[str], indent: str = " " * 4) -> List[str]:
    return [(f"{indent}{line}" if line else line) for line in lines]


def cleanup_str(text: str) -> str:
    return text.replace("\r", "").replace("\n", "")


def get_deprecated(text: Optional[str]) -> Optional[str]:
    if not text:
        return None

    lines = text.splitlines(keepends=False)
    for line in lines:
        if line.startswith("@deprecated"):
            return line.replace("@deprecated", "").strip()
    return None


def generate_extras(
    type_def: Union[
        model.Enum,
        model.EnumItem,
        model.Property,
        model.TypeAlias,
        model.Structure,
        model.Request,
        model.Notification,
    ]
) -> List[str]:
    deprecated = get_deprecated(type_def.documentation)
    extras = []
    if type_def.deprecated:
        extras += [f'[Obsolete("{cleanup_str(type_def.deprecated)}")]']
    elif deprecated:
        extras += [f'[Obsolete("{cleanup_str(deprecated)}")]']
    if type_def.proposed:
        extras += [f"[Proposed]"]
    if type_def.since:
        extras += [f'[Since("{cleanup_str(type_def.since)}")]']

    if hasattr(type_def, "messageDirection"):
        if type_def.since:
            extras += [
                f"[Direction(MessageDirection.{to_upper_camel_case(type_def.messageDirection)})]"
            ]

    return extras


def get_usings(types: List[str]) -> List[str]:
    usings = []

    for t in ["DataMember", "DataContract"]:
        if t in types:
            usings.append("using System.Runtime.Serialization;")

    for t in ["JsonConverter", "JsonConstructor", "JsonProperty", "NullValueHandling"]:
        if t in types:
            usings.append("using Newtonsoft.Json;")

    for t in ["JToken", "JObject", "JArray"]:
        if t in types:
            usings.append("using Newtonsoft.Json.Linq;")

    for t in ["List", "Dictionary"]:
        if t in types:
            usings.append("using System.Collections.Generic;")

    for t in ["ImmutableArray", "ImmutableDictionary"]:
        if t in types:
            usings.append("using System.Collections.Immutable;")

    return sorted(list(set(usings)))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import List


def license_header() -> List[str]:
    return [
        "Copyright (c) Microsoft Corporation. All rights reserved.",
        "Licensed under the MIT License.",
    ]


def package_description() -> List[str]:
    return ["Language Server Protocol types for Rust generated from LSP specification."]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import hashlib
import itertools
import json
import logging
import re
from copy import deepcopy
from typing import Any, Dict, List, Optional, Tuple, Union

import generator.model as model

LSP_MAX_INT = 2**31 - 1
LSP_MIN_INT = -(2**31)

LSP_MAX_UINT = 2**31 - 1
LSP_MIN_UINT = 0

LSP_OVER_MAX_INT = LSP_MAX_INT + 1
LSP_UNDER_MIN_INT = LSP_MIN_INT - 1

LSP_OVER_MAX_UINT = LSP_MAX_UINT + 1
LSP_UNDER_MIN_UINT = LSP_MIN_UINT - 1


def get_hash_from(text: str) -> str:
    return hashlib.sha256(text.encode("utf-8")).hexdigest()


def request_variants(method: str):
    for id_value in [1, LSP_MAX_INT, LSP_MIN_INT, "string-id-1"]:
        yield True, {"jsonrpc": "2.0", "id": id_value, "method": method}
    for id_value in [LSP_OVER_MAX_INT, LSP_UNDER_MIN_INT, 1.0, True, None]:
        yield False, {"jsonrpc": "2.0", "id": id_value, "method": method}
    yield False, {"jsonrpc": "2.0", "method": method}
    yield False, {"jsonrpc": "2.0", "id": 1}
    yield False, {"id": 1, "method": method}


def response_variants():
    for id_value in [1, LSP_MAX_INT, LSP_MIN_INT, "string-id-1"]:
        yield True, {"jsonrpc": "2.0", "id": id_value}
    for id_value in [LSP_OVER_MAX_INT, LSP_UNDER_MIN_INT, 1.0, True, None]:
        yield False, {"jsonrpc": "2.0", "id": id_value}
    yield False, {"jsonrpc": "2.0"}
    yield False, {"id": 1}


def notify_variants(method: str):
    yield True, {"jsonrpc": "2.0", "method": method}
    yield False, {"jsonrpc": "2.0", "id": 1, "method": method}


def _get_struct(name: str, spec: model.LSPModel) -> Optional[model.Structure]:
    for struct in spec.structures:
        if struct.name == name:
            return struct
    return None


def _get_type_alias(name: str, spec: model.LSPModel) -> Optional[model.TypeAlias]:
    for alias in spec.typeAliases:
        if alias.name == name:
            return alias
    return None


def _get_enum(name: str, spec: model.LSPModel) -> Optional[model.Enum]:
    for enum in spec.enumerations:
        if enum.name == name:
            return enum
    return None


class Ignore:
    pass


def extend(arr: List[Any], length: int) -> List[Any]:
    result = arr * (length // len(arr)) + arr[: length % len(arr)]
    return result


def extend_all(lists: List[List[Any]]) -> List[List[Any]]:
    max_len = max(len(l) for l in lists)
    max_len = min(1000, max_len)
    return [extend(l, max_len) for l in lists]


def has_null_base_type(items: List[model.LSP_TYPE_SPEC]) -> bool:
    return any(item.kind == "base" and item.name == "null" for item in items)


def filter_null_base_type(
    items: List[model.LSP_TYPE_SPEC],
) -> List[model.LSP_TYPE_SPEC]:
    return [item for item in items if not (item.kind == "base" and item.name == "null")]


def generate_for_base(name: str):
    if name == "string":
        yield (True, "some string 🐍🐜")
        yield (True, "")
    elif name == "integer":
        yield (True, 1)
        yield (True, LSP_MAX_INT)
        yield (True, LSP_MIN_INT)
        yield (False, LSP_OVER_MAX_INT)
        yield (False, LSP_UNDER_MIN_INT)
    elif name == "decimal":
        yield (True, 1.0)
    elif name == "boolean":
        yield (True, True)
        yield (True, False)
    elif name == "null":
        yield (True, None)
    elif name == "uinteger":
        yield (True, 1)
        yield (True, LSP_MAX_UINT)
        yield (True, LSP_MIN_UINT)
        yield (False, LSP_OVER_MAX_UINT)
        yield (False, LSP_UNDER_MIN_UINT)
    elif name in ["URI", "DocumentUri"]:
        yield (True, "file:///some/path")
    elif name == "RegExp":
        yield (True, ".*")


def generate_for_array(
    type_def: model.LSP_TYPE_SPEC, spec: model.LSPModel, visited: List[str]
):
    generated = list(generate_for_type(type_def, spec, visited))
    yield (True, [])

    # array with 1 item
    for valid, value in generated:
        if not isinstance(value, Ignore):
            yield (valid, [value])

    # array with 2 items
    generated = generated[:100] if len(generated) > 100 else generated
    values = itertools.product(generated, repeat=2)
    for (valid1, value1), (valid2, value2) in values:
        if not isinstance(value1, Ignore) and not isinstance(value2, Ignore):
            yield (valid1 and valid2, [value1, value2])


def generate_for_tuple(
    type_defs: List[model.LSP_TYPE_SPEC], spec: model.LSPModel, visited: List[str]
):
    generated = [
        list(generate_for_type(type_def, spec, visited)) for type_def in type_defs
    ]
    products = zip(*extend_all(generated))

    for product in products:
        is_valid = all(valid for valid, _ in product)
        values = [value for _, value in product if not isinstance(value, Ignore)]
        yield (is_valid, tuple(values))


def generate_for_map(
    key_def: model.LSP_TYPE_SPEC,
    value_def: model.LSP_TYPE_SPEC,
    spec: model.LSPModel,
    visited: List[str],
):
    key_values = list(generate_for_type(key_def, spec, visited))
    value_values = list(generate_for_type(value_def, spec, visited))

    for key_valid, key_value in key_values:
        for value_valid, value_value in value_values:
            if not (isinstance(key_value, Ignore) or isinstance(value_value, Ignore)):
                yield (key_valid and value_valid, {key_value: value_value})


def get_all_extends(struct_def: model.Structure, spec) -> List[model.Structure]:
    extends = []
    for extend in struct_def.extends:
        extends.append(_get_struct(extend.name, spec))
        for struct in get_all_extends(_get_struct(extend.name, spec), spec):
            if not any(struct.name == e.name for e in extends):
                extends.append(struct)
    return extends


def get_all_properties(struct: model.Structure, spec) -> List[model.Property]:
    properties = []
    for prop in struct.properties:
        properties.append(prop)

    for extend in get_all_extends(struct, spec):
        for prop in get_all_properties(extend, spec):
            if not any(prop.name == p.name for p in properties):
                properties.append(prop)

    if not all(mixin.kind == "reference" for mixin in struct.mixins):
        raise ValueError(f"Struct {struct.name} has non-reference mixins")
    for mixin in [_get_struct(mixin.name, spec) for mixin in struct.mixins]:
        for prop in get_all_properties(mixin, spec):
            if not any(prop.name == p.name for p in properties):
                properties.append(prop)

    return properties


def generate_for_property(
    prop: model.Property, spec: model.LSPModel, visited: List[str]
):
    if prop.optional:
        yield (True, Ignore())
    yield from generate_for_type(prop.type, spec, visited)


def generate_for_reference(
    refname: str,
    spec: model.LSPModel,
    visited: List[str],
):
    if len([name for name in visited if name == refname]) > 2:
        return

    ref = _get_struct(refname, spec)
    alias = _get_type_alias(refname, spec)
    enum = _get_enum(refname, spec)
    if ref:
        properties = get_all_properties(ref, spec)
        if properties:
            names = [prop.name for prop in properties]
            value_variants = [
                list(generate_for_property(prop, spec, visited)) for prop in properties
            ]

            products = zip(*extend_all(value_variants))
            for variants in products:
                is_valid = all(valid for valid, _ in variants)
                values = [value for _, value in variants]
                variant = {
                    name: value
                    for name, value in zip(names, values)
                    if not isinstance(value, Ignore)
                }
                yield (is_valid, variant)
        else:
            yield (True, {"lspExtension": "some value"})
            yield (True, dict())
    elif alias:
        if refname in ["LSPObject", "LSPAny", "LSPArray"]:
            yield from (
                (True, value)
                for _, value in generate_for_type(alias.type, spec, visited)
            )
        else:
            yield from generate_for_type(alias.type, spec, visited)
    elif enum:
        value = enum.values[0].value
        yield (True, value)
        if isinstance(value, int):
            yield (bool(enum.supportsCustomValues), 12345)
        elif isinstance(value, str):
            yield (bool(enum.supportsCustomValues), "testCustomValue")
    else:
        raise ValueError(f"Unknown reference {refname}")


def generate_for_or(
    type_defs: List[model.LSP_TYPE_SPEC],
    spec: model.LSPModel,
    visited: List[str],
):
    if has_null_base_type(type_defs):
        yield (True, None)

    subset = filter_null_base_type(type_defs)
    generated = [
        list(generate_for_type(type_def, spec, visited)) for type_def in subset
    ]
    for valid, value in itertools.chain(*generated):
        yield (valid, value)


def generate_for_and(
    type_defs: List[model.LSP_TYPE_SPEC],
    spec: model.LSPModel,
    visited: List[str],
):
    generated = [
        list(generate_for_type(type_def, spec, visited)) for type_def in type_defs
    ]
    products = zip(*extend_all(generated))
    for variants in products:
        is_valid = all(valid for valid, _ in variants)
        values = [value for _, value in variants]
        variant = {}
        for value in values:
            variant.update(value)
        yield (is_valid, variant)


def generate_for_literal(
    type_def: model.LiteralType,
    spec: model.LSPModel,
    visited: List[str],
):
    if type_def.value.properties:
        names = [prop.name for prop in type_def.value.properties]
        value_variants = [
            list(generate_for_type(prop.type, spec, visited))
            for prop in type_def.value.properties
        ]

        products = zip(*extend_all(value_variants))
        for variants in products:
            is_valid = all(valid for valid, _ in variants)
            values = [value for _, value in variants]
            variant = {name: value for name, value in zip(names, values)}
            yield (is_valid, variant)
    else:
        # Literal with no properties are a way to extend LSP spec
        # see: https://github.com/microsoft/vscode-languageserver-node/issues/997
        yield (True, {"lspExtension": "some value"})
        yield (True, dict())


def generate_for_type(
    type_def: model.LSP_TYPE_SPEC,
    spec: model.LSPModel,
    visited: List[str],
):
    if type_def is None:
        yield (True, None)
    elif type_def.kind == "base":
        yield from generate_for_base(type_def.name)
    elif type_def.kind == "array":
        yield from generate_for_array(type_def.element, spec, visited)
    elif type_def.kind == "reference":
        yield from generate_for_reference(
            type_def.name, spec, visited + [type_def.name]
        )
    elif type_def.kind == "stringLiteral":
        yield (True, type_def.value)
        # yield (False, f"invalid@{type_def.value}")
    elif type_def.kind == "tuple":
        yield from generate_for_tuple(type_def.items, spec, visited)
    elif type_def.kind == "or":
        yield from generate_for_or(type_def.items, spec, visited)
    elif type_def.kind == "and":
        yield from generate_for_and(type_def.items, spec, visited)
    elif type_def.kind == "literal":
        yield from generate_for_literal(type_def, spec, visited)
    elif type_def.kind == "map":
        yield from generate_for_map(type_def.key, type_def.value, spec, visited)


def generate_requests(request: model.Request, spec: model.LSPModel):
    variants = zip(
        *extend_all(
            [
                list(request_variants(request.method)),
                list(generate_for_type(request.params, spec, [])),
            ]
        )
    )

    for (valid1, base), (valid2, params) in variants:
        valid = valid1 and valid2
        if isinstance(params, Ignore):
            yield (valid, base)
        else:
            message = deepcopy(base)
            message.update({"params": params})
            yield (valid1 and valid2, message)


def generate_notifications(notify: model.Notification, spec: model.LSPModel) -> None:
    variants = zip(
        *extend_all(
            [
                list(notify_variants(notify.method)),
                list(generate_for_type(notify.params, spec, [])),
            ]
        )
    )

    for (valid1, base), (valid2, params) in variants:
        valid = valid1 and valid2
        if isinstance(params, Ignore):
            yield (valid, base)
        else:
            message = deepcopy(base)
            message.update({"params": params})
            yield (valid1 and valid2, message)


RESPONSE_ERROR = model.Structure(
    **{
        "name": "ResponseError",
        "properties": [
            {
                "name": "code",
                "type": {"kind": "base", "name": "integer"},
            },
            {
                "name": "message",
                "type": {"kind": "base", "name": "string"},
            },
            {
                "name": "data",
                "type": {"kind": "reference", "name": "LSPObject"},
                "optional": True,
            },
        ],
    }
)


def generate_responses(request: model.Request, spec: model.LSPModel) -> None:
    variants = zip(
        *extend_all(
            [
                list(response_variants()),
                list(generate_for_type(request.result, spec, [])),
                list(
                    generate_for_type(
                        model.ReferenceType("reference", "ResponseError"), spec, []
                    )
                ),
            ]
        )
    )

    for (valid1, base), (valid2, result), (valid3, error) in variants:
        valid = valid1 and valid2 and valid3
        if isinstance(result, Ignore):
            yield (valid, base)
        else:
            message = deepcopy(base)
            message.update({"result": result})
            message.update({"error": error})
            yield (valid, message)


PARTS_RE = re.compile(r"(([a-z0-9])([A-Z]))")


def get_parts(name: str) -> List[str]:
    name = name.replace("_", " ")
    return PARTS_RE.sub(r"\2 \3", name).split()


def to_upper_camel_case(name: str) -> str:
    return "".join([c.capitalize() for c in get_parts(name)])


def lsp_method_to_name(method: str) -> str:
    if method.startswith("$"):
        method = method[1:]
    method = method.replace("/", "_")
    return to_upper_camel_case(method)


def generate(spec: model.LSPModel, logger: logging.Logger):
    spec.structures.append(RESPONSE_ERROR)
    testdata = {}
    for request in spec.requests:
        counter = 0
        for valid, value in generate_requests(request, spec):
            content = json.dumps(value, indent=4, ensure_ascii=False)
            name = f"{lsp_method_to_name(request.method)}Request-{valid}-{get_hash_from(content)}.json"
            if name in testdata:
                continue
            testdata[name] = content
            counter += 1
        logger.info(f"Generated {counter} variants for Request: {request.method}")

        for valid, value in generate_responses(request, spec):
            content = json.dumps(value, indent=4, ensure_ascii=False)
            name = f"{lsp_method_to_name(request.method)}Response-{valid}-{get_hash_from(content)}.json"
            if name in testdata:
                continue
            testdata[name] = content
            counter += 1
        logger.info(f"Generated {counter} variants for Response: {request.method}")

    for notify in spec.notifications:
        counter = 0
        for valid, value in generate_notifications(notify, spec):
            content = json.dumps(value, indent=4, ensure_ascii=False)
            name = f"{lsp_method_to_name(notify.method)}Notification-{valid}-{get_hash_from(content)}.json"
            if name in testdata:
                continue
            testdata[name] = content
            counter += 1
        logger.info(f"Generated {counter} variants for Notification: {notify.method}")

    logger.info(f"Generated {len(testdata)} test variants")
    return testdata



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from __future__ import annotations

import uuid
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar, Union

import attrs

LSP_TYPE_SPEC = Union[
    "BaseType",
    "ReferenceType",
    "OrType",
    "AndType",
    "ArrayType",
    "LiteralType",
    "StringLiteralType",
    "MapType",
    "TupleType",
]


def partial_apply(callable):
    def apply(x):
        if isinstance(x, dict):
            return callable(**x)
        else:
            return x

    return apply


def list_converter(callable):
    def apply(x):
        if isinstance(x, dict):
            return callable(**x)
        else:
            return x

    def converter(x: List[Any]) -> List[Any]:
        return list(map(apply, x))

    return converter


def enum_validator(instance: Enum, attribute: Any, value: Any) -> None:
    test_type = str if instance.type.name == "string" else int
    for e in value:
        if not isinstance(e.value, test_type):
            raise ValueError(
                f"Value of {instance.name}.{e.name} is not of type {test_type}: {e.value}."
            )


def convert_to_lsp_type(**type_info) -> Optional[LSP_TYPE_SPEC]:
    if type_info is None:
        return None
    lut: Dict[str, LSP_TYPE_SPEC] = {
        "base": BaseType,
        "reference": ReferenceType,
        "array": ArrayType,
        "or": OrType,
        "and": AndType,
        "literal": LiteralType,
        "map": MapType,
        "stringLiteral": StringLiteralType,
        "tuple": TupleType,
    }
    callable = lut.get(type_info["kind"])
    if callable:
        return callable(**type_info)

    raise ValueError(f"Unknown LSP type: {type_info}")


def type_validator(instance: Any, attribute: str, value: Any) -> bool:
    return isinstance(
        value,
        (
            BaseType,
            ReferenceType,
            ArrayType,
            OrType,
            LiteralType,
            AndType,
            MapType,
            StringLiteralType,
            TupleType,
        ),
    )


@attrs.define
class EnumItem:
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    value: Union[str, int] = attrs.field(
        validator=attrs.validators.instance_of((str, int))
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, EnumItem):
            return self.name == other.name and self.value == other.value
        return False

    def get_inner_types(self) -> List[Type]:
        return []


@attrs.define
class EnumValueType:
    kind: str = attrs.field(validator=attrs.validators.in_(["base"]))
    name: str = attrs.field(
        validator=attrs.validators.in_(["string", "integer", "uinteger"])
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, EnumValueType):
            return self.name == other.name and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        return []


@attrs.define
class Enum:
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: EnumValueType = attrs.field(converter=partial_apply(EnumValueType))
    values: Iterable[EnumItem] = attrs.field(
        validator=enum_validator,
        converter=list_converter(EnumItem),
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    supportsCustomValues: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Enum):
            return (
                self.name == other.name
                and self.type == other.type
                and self.values == other.values
            )
        return False

    def get_inner_types(self) -> List[Type]:
        types = [self.type]
        for value in self.values:
            types.append(value)
            types.extend(value.get_inner_types())
        return types


@attrs.define
class BaseType:
    kind: str = attrs.field(validator=attrs.validators.in_(["base"]))
    name: str = attrs.field(
        validator=attrs.validators.in_(
            [
                "URI",
                "DocumentUri",
                "integer",
                "uinteger",
                "decimal",
                "RegExp",
                "string",
                "boolean",
                "null",
            ]
        )
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BaseType):
            return self.name == other.name and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        return []


@attrs.define
class ReferenceType:
    kind: str = attrs.field(validator=attrs.validators.in_(["reference"]))
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ReferenceType):
            return self.name == other.name and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        return []


@attrs.define
class StringLiteralType:
    kind: str = attrs.field(validator=attrs.validators.in_(["stringLiteral"]))
    value: str = attrs.field(validator=attrs.validators.instance_of(str))
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, StringLiteralType):
            return self.value == other.value and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        return []


@attrs.define
class OrType:
    kind: str = attrs.field(validator=attrs.validators.in_(["or"]))
    items: Iterable[LSP_TYPE_SPEC] = attrs.field(
        converter=list_converter(convert_to_lsp_type)
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, OrType):
            return self.items == other.items and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        types = []
        for item in self.items:
            types.append(item)
            types.extend(item.get_inner_types())
        return types


@attrs.define
class AndType:
    kind: str = attrs.field(validator=attrs.validators.in_(["and"]))
    items: Iterable[LSP_TYPE_SPEC] = attrs.field(
        converter=list_converter(convert_to_lsp_type),
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, AndType):
            return self.items == other.items and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        types = []
        for item in self.items:
            types.append(item)
            types.extend(item.get_inner_types())
        return types


@attrs.define
class ArrayType:
    kind: str = attrs.field(validator=attrs.validators.in_(["array"]))
    element: LSP_TYPE_SPEC = attrs.field(
        validator=type_validator, converter=partial_apply(convert_to_lsp_type)
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ArrayType):
            return self.element == other.element and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        return [self.element] + self.element.get_inner_types()


@attrs.define
class TupleType:
    kind: str = attrs.field(validator=attrs.validators.in_(["tuple"]))
    items: Iterable[LSP_TYPE_SPEC] = attrs.field(
        converter=list_converter(convert_to_lsp_type)
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TupleType):
            return self.items == other.items and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        types = []
        for item in self.items:
            types.append(item)
            types.extend(item.get_inner_types())
        return types


@attrs.define
class BaseMapKeyType:
    kind: str = attrs.field(validator=attrs.validators.in_(["base"]))
    name: str = attrs.field(
        validator=attrs.validators.in_(
            [
                "URI",
                "DocumentUri",
                "integer",
                "string",
            ]
        )
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BaseMapKeyType):
            return self.name == other.name and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        return []


@attrs.define
class ReferenceMapKeyType:
    kind: str = attrs.field(validator=attrs.validators.in_(["reference"]))
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ReferenceMapKeyType):
            return self.name == other.name and self.kind == other.kind
        return False

    def get_inner_types(self) -> List[Type]:
        return []


def convert_map_key(
    key_info: Dict[str, Any]
) -> Union[BaseMapKeyType, ReferenceMapKeyType]:
    if key_info["kind"] == "base":
        return BaseMapKeyType(**key_info)
    return ReferenceMapKeyType(**key_info)


@attrs.define
class MapType:
    kind: str = attrs.field(validator=attrs.validators.in_(["map"]))
    key: Union[BaseMapKeyType, ReferenceMapKeyType] = attrs.field(
        converter=convert_map_key
    )
    value: LSP_TYPE_SPEC = attrs.field(
        validator=type_validator, converter=partial_apply(convert_to_lsp_type)
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, MapType):
            return (
                self.key == other.key
                and self.value == other.value
                and self.kind == other.kind
            )
        return False

    def get_inner_types(self) -> List[Type]:
        return (
            [self.value, self.key]
            + self.value.get_inner_types()
            + self.key.get_inner_types()
        )


@attrs.define
class MetaData:
    version: str = attrs.field(validator=attrs.validators.instance_of(str))
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, MetaData):
            return self.version == other.version
        return False

    def get_inner_types(self) -> List[Type]:
        return []


@attrs.define
class Property:
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: LSP_TYPE_SPEC = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
    )
    optional: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Property):
            return (
                self.name == other.name
                and self.type == other.type
                and self.optional == other.optional
            )
        return False

    def get_inner_types(self) -> List[Type]:
        return [self.type] + self.type.get_inner_types()


@attrs.define
class LiteralValue:
    properties: Iterable[Property] = attrs.field(
        converter=list_converter(Property),
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LiteralValue):
            return self.properties == other.properties
        return False

    def get_inner_types(self) -> List[Type]:
        types = []
        for prop in self.properties:
            types.append(prop)
            types.extend(prop.get_inner_types())
        return types


@attrs.define
class LiteralType:
    kind: str = attrs.field(validator=attrs.validators.in_(["literal"]))
    value: LiteralValue = attrs.field(
        validator=attrs.validators.instance_of(LiteralValue),
        converter=partial_apply(LiteralValue),
    )
    name: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LiteralType):
            return (
                self.value == other.value
                and self.kind == other.kind
                and self.name == other.name
            )
        return False

    def get_inner_types(self) -> List[Type]:
        return [self.value] + self.value.get_inner_types()


@attrs.define
class TypeAlias:
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    type: LSP_TYPE_SPEC = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, TypeAlias):
            return (
                self.name == other.name
                and self.type == other.type
                and self.kind == other.kind
            )
        return False

    def get_inner_types(self) -> List[Type]:
        return [self.type] + self.type.get_inner_types()


@attrs.define
class Structure:
    name: str = attrs.field(validator=attrs.validators.instance_of(str))
    properties: Iterable[Property] = attrs.field(
        converter=list_converter(Property),
    )
    extends: Optional[Iterable[LSP_TYPE_SPEC]] = attrs.field(
        converter=list_converter(convert_to_lsp_type), default=[]
    )
    mixins: Optional[Iterable[LSP_TYPE_SPEC]] = attrs.field(
        converter=list_converter(convert_to_lsp_type), default=[]
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Structure):
            return (
                self.name == other.name
                and self.properties == other.properties
                and self.extends == other.extends
                and self.mixins == other.mixins
            )
        return False

    def get_inner_types(self) -> List[Type]:
        types = []
        for prop in self.properties:
            types.append(prop.type)
            types.extend(prop.type.get_inner_types())
        for ext in self.extends:
            types.append(ext)
            types.extend(ext.get_inner_types())
        for mixin in self.mixins:
            types.append(mixin)
            types.extend(mixin.get_inner_types())
        return types


@attrs.define
class Notification:
    method: str = attrs.field(validator=attrs.validators.instance_of(str))
    messageDirection: str = attrs.field(
        validator=attrs.validators.in_(["clientToServer", "serverToClient", "both"])
    )
    params: Optional[LSP_TYPE_SPEC] = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
        default=None,
    )
    registrationOptions: Optional[LSP_TYPE_SPEC] = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
        default=None,
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    registrationMethod: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Notification):
            return (
                self.method == other.method
                and self.messageDirection == other.messageDirection
                and self.params == other.params
                and self.registrationOptions == other.registrationOptions
                and self.registrationMethod == other.registrationMethod
            )
        return False

    def get_inner_types(self) -> List[Type]:
        types = []
        if self.params:
            types.append(self.params)
            types.extend(self.params.get_inner_types())
        if self.registrationOptions:
            types.append(self.registrationOptions)
            types.extend(self.registrationOptions.get_inner_types())
        return types


@attrs.define
class Request:
    method: str = attrs.field(validator=attrs.validators.instance_of(str))
    messageDirection: str = attrs.field(
        validator=attrs.validators.in_(["clientToServer", "serverToClient", "both"])
    )
    params: Optional[LSP_TYPE_SPEC] = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
        default=None,
    )
    result: Optional[LSP_TYPE_SPEC] = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
        default=None,
    )
    partialResult: Optional[LSP_TYPE_SPEC] = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
        default=None,
    )
    errorData: Optional[LSP_TYPE_SPEC] = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
        default=None,
    )
    registrationOptions: Optional[LSP_TYPE_SPEC] = attrs.field(
        validator=type_validator,
        converter=partial_apply(convert_to_lsp_type),
        default=None,
    )
    proposed: Optional[bool] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(bool)),
        default=None,
    )
    documentation: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    since: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    registrationMethod: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    deprecated: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )
    id_: Optional[str] = attrs.field(
        converter=lambda x: str(uuid.uuid4()),
        validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        default=None,
    )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Request):
            return (
                self.method == other.method
                and self.messageDirection == other.messageDirection
                and self.params == other.params
                and self.result == other.result
                and self.partialResult == other.partialResult
                and self.errorData == other.errorData
                and self.registrationOptions == other.registrationOptions
                and self.registrationMethod == other.registrationMethod
            )
        return False

    def get_inner_types(self) -> List[Type]:
        types = []
        if self.params:
            types.append(self.params)
            types.extend(self.params.get_inner_types())
        if self.result:
            types.append(self.result)
            types.extend(self.result.get_inner_types())
        if self.partialResult:
            types.append(self.partialResult)
            types.extend(self.partialResult.get_inner_types())
        if self.errorData:
            types.append(self.errorData)
            types.extend(self.errorData.get_inner_types())
        if self.registrationOptions:
            types.append(self.registrationOptions)
            types.extend(self.registrationOptions.get_inner_types())
        return types


@attrs.define
class LSPModel:
    requests: Iterable[Request] = attrs.field(converter=list_converter(Request))
    notifications: Iterable[Notification] = attrs.field(
        converter=list_converter(Notification)
    )
    structures: Iterable[Structure] = attrs.field(converter=list_converter(Structure))
    enumerations: Iterable[Enum] = attrs.field(converter=list_converter(Enum))
    typeAliases: Iterable[TypeAlias] = attrs.field(converter=list_converter(TypeAlias))
    metaData: MetaData = attrs.field(converter=lambda x: MetaData(**x))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LSPModel):
            return (
                self.requests == other.requests
                and self.notifications == other.notifications
                and self.structures == other.structures
                and self.enumerations == other.enumerations
                and self.typeAliases == other.typeAliases
                and self.metaData == other.metaData
            )
        return False

    def get_inner_types(self) -> List[Type]:
        types = self.metaData.get_inner_types()
        for r in self.requests:
            types.extend(r.get_inner_types())
        for n in self.notifications:
            types.extend(n.get_inner_types())
        for s in self.structures:
            types.extend(s.get_inner_types())
        for e in self.enumerations:
            types.extend(e.get_inner_types())
        for t in self.typeAliases:
            types.extend(t.get_inner_types())
        return types


def create_lsp_model(models: List[Dict[str, Any]]) -> LSPModel:
    if len(models) >= 1:
        spec = LSPModel(**models[0])

    if len(models) >= 2:
        for model in models[1:]:
            addition = LSPModel(**model)
            spec.requests.extend(addition.requests)
            spec.notifications.extend(addition.notifications)
            spec.structures.extend(addition.structures)
            spec.enumerations.extend(addition.enumerations)
            spec.typeAliases.extend(addition.typeAliases)

    return spec



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.


import re
from typing import List

BASIC_LINK_RE = re.compile(r"{@link +(\w+) ([\w ]+)}")
BASIC_LINK_RE2 = re.compile(r"{@link +(\w+)\.(\w+) ([\w \.`]+)}")
BASIC_LINK_RE3 = re.compile(r"{@link +(\w+)}")
BASIC_LINK_RE4 = re.compile(r"{@link +(\w+)\.(\w+)}")
PARTS_RE = re.compile(r"(([a-z0-9])([A-Z]))")
DEFAULT_INDENT = "    "


def lines_to_comments(lines: List[str]) -> List[str]:
    return ["// " + line for line in lines]


def lines_to_doc_comments(lines: List[str]) -> List[str]:
    doc = []
    for line in lines:
        line = BASIC_LINK_RE.sub(r"[\2][\1]", line)
        line = BASIC_LINK_RE2.sub(r"[\3][`\1::\2`]", line)
        line = BASIC_LINK_RE3.sub(r"[\1]", line)
        line = BASIC_LINK_RE4.sub(r"[`\1::\2`]", line)
        if line.startswith("///"):
            doc.append(line)
        else:
            doc.append("/// " + line)
    return doc


def lines_to_block_comment(lines: List[str]) -> List[str]:
    return ["/*"] + lines + ["*/"]


def get_parts(name: str) -> List[str]:
    name = name.replace("_", " ")
    return PARTS_RE.sub(r"\2 \3", name).split()


def to_snake_case(name: str) -> str:
    return "_".join([part.lower() for part in get_parts(name)])


def has_upper_case(name: str) -> bool:
    return any(c.isupper() for c in name)


def is_snake_case(name: str) -> bool:
    return (
        not name.startswith("_")
        and not name.endswith("_")
        and ("_" in name)
        and not has_upper_case(name)
    )


def to_upper_camel_case(name: str) -> str:
    return "".join([c.capitalize() for c in get_parts(name)])


def to_camel_case(name: str) -> str:
    parts = get_parts(name)
    if len(parts) > 1:
        return parts[0] + "".join([c.capitalize() for c in parts[1:]])
    else:
        return parts[0]


def indent_lines(lines: List[str], indent: str = DEFAULT_INDENT) -> List[str]:
    return [f"{indent}{line}" for line in lines]



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Dict, Iterable, List, Optional

import generator.model as model

from .rust_commons import (
    TypeData,
    fix_lsp_method_name,
    generate_extras,
    generate_literal_struct_name,
    generate_property,
    get_extended_properties,
    get_message_type_name,
    get_type_name,
    struct_wrapper,
    type_alias_wrapper,
)
from .rust_lang_utils import get_parts, lines_to_doc_comments, to_upper_camel_case


def generate_type_aliases(spec: model.LSPModel, types: TypeData) -> None:
    for alias in spec.typeAliases:
        if not types.has_id(alias):
            generate_type_alias(alias, types, spec)


def _get_doc(doc: Optional[str]) -> str:
    if doc:
        return lines_to_doc_comments(doc.splitlines(keepends=False))
    return []


def _is_some_array_type(items: Iterable[model.LSP_TYPE_SPEC]) -> bool:
    items_list = list(items)
    assert len(items_list) == 2
    item1, item2 = items_list

    if item1.kind == "array" and item2.kind == "reference":
        return item1.element.kind == "reference" and item1.element.name == item2.name

    if item2.kind == "array" and item1.kind == "reference":
        return item2.element.kind == "reference" and item2.element.name == item1.name
    return False


def _get_some_array_code(
    items: Iterable[model.LSP_TYPE_SPEC],
    types: Dict[str, List[str]],
    spec: model.LSPModel,
) -> List[str]:
    assert _is_some_array_type(items)
    items_list = list(items)
    item1 = items_list[0]
    item2 = items_list[1]

    if item1.kind == "array" and item2.kind == "reference":
        return [
            f"    One({get_type_name(item2, types, spec)}),",
            f"    Many({get_type_name(item1, types, spec)}),",
        ]

    if item2.kind == "array" and item1.kind == "reference":
        return [
            f"    One({get_type_name(item1, types, spec)}),",
            f"    Many({get_type_name(item2, types, spec)}),",
        ]
    return []


def _get_common_name(items: Iterable[model.LSP_TYPE_SPEC], kind: str) -> List[str]:
    names = [get_parts(item.name) for item in list(items) if item.kind == kind]
    if len(names) < 2:
        return []

    smallest = min(names, key=len)
    common = []
    for i in range(len(smallest)):
        if all(name[i] == smallest[i] for name in names):
            common.append(smallest[i])
    return common


def _is_all_reference_similar_type(alias: model.TypeAlias) -> bool:
    items_list = list(alias.type.items)
    return all(item.kind in ["reference", "base", "literal"] for item in items_list)


def _get_all_reference_similar_code(
    alias: model.TypeAlias,
    types: TypeData,
    spec: model.LSPModel,
) -> List[str]:
    items = alias.type.items
    assert _is_all_reference_similar_type(alias)

    # Ensure all literal types have a name
    for item in list(items):
        if item.kind == "literal":
            get_type_name(item, types, spec, None, alias.name)

    common_name = [
        i.lower()
        for i in (
            _get_common_name(items, "reference")
            + _get_common_name(items, "literal")
            + ["struct"]
        )
    ]

    lines = []
    value = 0
    field_names = []
    for item in list(items):
        if item.kind == "base" and item.name == "null":
            lines += ["None,"]
            field_names += ["None"]
        elif item.kind == "base":
            name = _base_to_field_name(item.name)
            lines += [f"{name}({get_type_name(item, types, spec)}),"]
            field_names += [name]
        elif item.kind == "reference":
            name = [
                part for part in get_parts(item.name) if part.lower() not in common_name
            ]
            if len(name) == 0:
                name = [f"Value{value}"]
                value += 1
            common_name += [n.lower() for n in name]
            name = to_upper_camel_case("".join(name))
            field_names += [name]
            lines += [f"{name}({get_type_name(item, types, spec)}),"]
        elif item.kind == "literal":
            name = [
                part for part in get_parts(item.name) if part.lower() not in common_name
            ]
            optional_props = [p for p in item.value.properties if p.optional]
            required_props = [p for p in item.value.properties if not p.optional]

            # Try picking a name using required props first and then optional props
            if len(name) == 0:
                for p in required_props + optional_props:
                    name = [
                        part
                        for part in get_parts(p.name)
                        if part.lower() not in common_name
                    ]
                    if len(name) != 0:
                        break

            # If we still don't have a name, then try picking a name using required props
            # and then optional props without checking for common name list. But check
            # that the name is not already used.
            if len(name) == 0:
                for p in required_props + optional_props:
                    if to_upper_camel_case(p.name) not in field_names:
                        name = get_parts(p.name)
                        break

            # If we still don't have a name, then just use a generic "Value{int}" as name
            if len(name) == 0:
                name = [f"Value{value}"]
                value += 1
            common_name += [n.lower() for n in name]
            name = to_upper_camel_case("".join(name))
            field_names += [name]
            lines += [f"{name}({item.name}),"]
        else:
            raise ValueError(f"Unknown type {item}")
    return lines


def _base_to_field_name(base_name: str) -> str:
    if base_name == "boolean":
        return "Bool"
    if base_name == "integer":
        return "Int"
    if base_name == "decimal":
        return "Real"
    if base_name == "string":
        return "String"
    if base_name == "uinteger":
        return "UInt"
    if base_name == "null":
        return "None"
    raise ValueError(f"Unknown base type {base_name}")


def _get_literal_field_name(literal: model.LiteralType, types: TypeData) -> str:
    properties = list(literal.value.properties)

    if len(properties) == 1 and properties[0].kind == "base":
        return _base_to_field_name(properties[0].name)

    if len(properties) == 1 and properties[0].kind == "reference":
        return to_upper_camel_case(properties[0].name)

    return generate_literal_struct_name(literal, types)


def _generate_or_type_alias(
    alias_def: model.TypeAlias, types: Dict[str, List[str]], spec: model.LSPModel
) -> List[str]:
    inner = []

    if len(alias_def.type.items) == 2 and _is_some_array_type(alias_def.type.items):
        inner += _get_some_array_code(alias_def.type.items, types, spec)
    elif _is_all_reference_similar_type(alias_def):
        inner += _get_all_reference_similar_code(alias_def, types, spec)
    else:
        index = 0

        for sub_type in alias_def.type.items:
            if sub_type.kind == "base" and sub_type.name == "null":
                inner += [f"None,"]
            else:
                inner += [f"ValueType{index}({get_type_name(sub_type, types, spec)}),"]
            index += 1
    return type_alias_wrapper(alias_def, inner)


def generate_type_alias(
    alias_def: model.TypeAlias, types: TypeData, spec: model.LSPModel
) -> List[str]:
    doc = _get_doc(alias_def.documentation)
    doc += generate_extras(alias_def)

    lines = []
    if alias_def.type.kind == "reference":
        lines += doc
        lines += [f"pub type {alias_def.name} = {alias_def.type.name};"]
    elif alias_def.type.kind == "array":
        lines += doc
        lines += [
            f"pub type {alias_def.name} = {get_type_name(alias_def.type, types, spec)};"
        ]
    elif alias_def.type.kind == "or":
        lines += _generate_or_type_alias(alias_def, types, spec)
    elif alias_def.type.kind == "and":
        raise ValueError("And type not supported")
    elif alias_def.type.kind == "literal":
        lines += doc
        lines += [
            f"pub type {alias_def.name} = {get_type_name(alias_def.type, types, spec)};"
        ]
    elif alias_def.type.kind == "base":
        lines += doc
        lines += [
            f"pub type {alias_def.name} = {get_type_name(alias_def.type, types, spec)};"
        ]
    else:
        pass

    types.add_type_info(alias_def, alias_def.name, lines)


def generate_structures(spec: model.LSPModel, types: TypeData) -> Dict[str, List[str]]:
    for struct in spec.structures:
        if not types.has_id(struct):
            generate_struct(struct, types, spec)
    return types


def generate_struct(
    struct_def: model.Structure, types: TypeData, spec: model.LSPModel
) -> None:
    inner = []
    for prop_def in get_extended_properties(struct_def, spec):
        inner += generate_property(prop_def, types, spec)

    lines = struct_wrapper(struct_def, inner)
    types.add_type_info(struct_def, struct_def.name, lines)


def generate_notifications(
    spec: model.LSPModel, types: TypeData
) -> Dict[str, List[str]]:
    for notification in spec.notifications:
        if not types.has_id(notification):
            generate_notification(notification, types, spec)
    return types


def required_rpc_properties(name: Optional[str] = None) -> List[model.Property]:
    props = [
        model.Property(
            name="jsonrpc",
            type=model.BaseType(kind="base", name="string"),
            optional=False,
            documentation="The version of the JSON RPC protocol.",
        ),
    ]
    if name:
        props += [
            model.Property(
                name="method",
                type=model.ReferenceType(kind="reference", name=name),
                optional=False,
                documentation="The method to be invoked.",
            ),
        ]
    return props


def generate_notification(
    notification_def: model.Notification, types: TypeData, spec: model.LSPModel
) -> None:
    properties = required_rpc_properties("LSPNotificationMethods")
    if notification_def.params:
        properties += [
            model.Property(
                name="params",
                type=notification_def.params,
            )
        ]

    inner = []
    for prop_def in properties:
        inner += generate_property(prop_def, types, spec)

    lines = struct_wrapper(notification_def, inner)
    types.add_type_info(
        notification_def, get_message_type_name(notification_def), lines
    )


def generate_required_request_types(
    spec: model.LSPModel, types: TypeData
) -> Dict[str, List[str]]:
    lsp_id = model.TypeAlias(
        name="LSPId",
        documentation="An identifier to denote a specific request.",
        type=model.OrType(
            kind="or",
            items=[
                model.BaseType(kind="base", name="integer"),
                model.BaseType(kind="base", name="string"),
            ],
        ),
    )
    generate_type_alias(lsp_id, types, spec)

    lsp_id_optional = model.TypeAlias(
        name="LSPIdOptional",
        documentation="An identifier to denote a specific response.",
        type=model.OrType(
            kind="or",
            items=[
                model.BaseType(kind="base", name="integer"),
                model.BaseType(kind="base", name="string"),
                model.BaseType(kind="base", name="null"),
            ],
        ),
    )
    generate_type_alias(lsp_id_optional, types, spec)


def generate_requests(spec: model.LSPModel, types: TypeData) -> Dict[str, List[str]]:
    generate_required_request_types(spec, types)
    for request in spec.requests:
        if not types.has_id(request):
            generate_request(request, types, spec)
            generate_response(request, types, spec)
            generate_partial_result(request, types, spec)
            generate_registration_options(request, types, spec)
    return types


def generate_request(
    request_def: model.Request, types: TypeData, spec: model.LSPModel
) -> None:
    properties = required_rpc_properties("LSPRequestMethods")

    properties += [
        model.Property(
            name="id",
            type=model.ReferenceType(kind="reference", name="LSPId"),
            optional=False,
            documentation="The request id.",
        )
    ]
    if request_def.params:
        properties += [
            model.Property(
                name="params",
                type=request_def.params,
            )
        ]

    inner = []
    for prop_def in properties:
        inner += generate_property(prop_def, types, spec)

    lines = struct_wrapper(request_def, inner)
    types.add_type_info(request_def, get_message_type_name(request_def), lines)


def generate_response(
    request_def: model.Request, types: TypeData, spec: model.LSPModel
) -> None:
    properties = required_rpc_properties("LSPRequestMethods")
    properties += [
        model.Property(
            name="id",
            type=model.ReferenceType(kind="reference", name="LSPIdOptional"),
            optional=False,
            documentation="The request id.",
        )
    ]
    if request_def.result:
        if request_def.result.kind == "base" and request_def.result.name == "null":
            properties += [
                model.Property(
                    name="result",
                    type=model.ReferenceType(kind="reference", name="LSPNull"),
                )
            ]
        else:
            properties += [
                model.Property(
                    name="result",
                    type=request_def.result,
                )
            ]
    name = fix_lsp_method_name(request_def.method)
    response_def = model.Structure(
        name=f"{name}Response",
        documentation=f"Response to the [{name}Request].",
        properties=properties,
        since=request_def.since,
        deprecated=request_def.deprecated,
    )

    inner = []
    for prop_def in properties:
        inner += generate_property(prop_def, types, spec)

    lines = struct_wrapper(response_def, inner)
    types.add_type_info(response_def, response_def.name, lines)


def generate_partial_result(
    request_def: model.Request, types: TypeData, spec: model.LSPModel
) -> None:
    if not request_def.partialResult:
        return

    if request_def.partialResult.kind not in ["and", "or"]:
        return


def generate_registration_options(
    request_def: model.Request, types: TypeData, spec: model.LSPModel
) -> None:
    if not request_def.registrationOptions:
        return

    if request_def.registrationOptions.kind not in ["and", "or"]:
        return



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import re
from typing import Any, Dict, List, Optional, Tuple, Union

import cattrs

from generator import model

from .dotnet_commons import TypeData
from .dotnet_constants import NAMESPACE
from .dotnet_helpers import (
    class_wrapper,
    generate_extras,
    get_doc,
    get_special_case_class_name,
    get_special_case_property_name,
    get_usings,
    indent_lines,
    lsp_method_to_name,
    namespace_wrapper,
    to_camel_case,
    to_upper_camel_case,
)

ORTYPE_CONVERTER_RE = re.compile(r"OrType<(?P<parts>.*)>")
IMMUTABLE_ARRAY_CONVERTER_RE = re.compile(r"ImmutableArray<(?P<elements>.*)>")


def _get_enum(name: str, spec: model.LSPModel) -> Optional[model.Enum]:
    for enum in spec.enumerations:
        if enum.name == name:
            return enum
    return None


def _get_struct(name: str, spec: model.LSPModel) -> Optional[model.Structure]:
    for struct in spec.structures:
        if struct.name == name:
            return struct
    return None


def _is_str_enum(enum_def: model.Enum) -> bool:
    return all(isinstance(item.value, str) for item in enum_def.values)


def _is_int_enum(enum_def: model.Enum) -> bool:
    return all(isinstance(item.value, int) for item in enum_def.values)


def lsp_to_base_types(lsp_type: model.BaseType):
    if lsp_type.name in ["string", "RegExp"]:
        return "string"
    elif lsp_type.name in ["DocumentUri", "URI"]:
        return "Uri"
    elif lsp_type.name in ["decimal"]:
        return "float"
    elif lsp_type.name in ["integer"]:
        return "int"
    elif lsp_type.name in ["uinteger"]:
        return "uint"
    elif lsp_type.name in ["boolean"]:
        return "bool"
    elif lsp_type.name in ["null"]:
        return "object"

    # null should be handled by the caller as an Option<> type
    raise ValueError(f"Unknown base type: {lsp_type.name}")


def get_types_for_usings(code: List[str]) -> List[str]:
    immutable = []
    for line in code:
        if "ImmutableArray<" in line:
            immutable.append("ImmutableArray")
        if "ImmutableDictionary<" in line:
            immutable.append("ImmutableDictionary")
    return list(set(immutable))


def has_null_base_type(items: List[model.LSP_TYPE_SPEC]) -> bool:
    return any(item.kind == "base" and item.name == "null" for item in items)


def filter_null_base_type(
    items: List[model.LSP_TYPE_SPEC],
) -> List[model.LSP_TYPE_SPEC]:
    return [item for item in items if not (item.kind == "base" and item.name == "null")]


def get_type_name(
    type_def: model.LSP_TYPE_SPEC,
    types: TypeData,
    spec: model.LSPModel,
    name_context: Optional[str] = None,
) -> str:
    name = None
    if type_def.kind == "reference":
        enum_def = _get_enum(type_def.name, spec)
        if enum_def and enum_def.supportsCustomValues:
            if _is_str_enum(enum_def):
                name = f"string"
            elif _is_int_enum(enum_def):
                name = f"int"
        else:
            name = get_special_case_class_name(type_def.name)
    elif type_def.kind == "array":
        name = f"ImmutableArray<{get_type_name(type_def.element, types, spec, name_context)}>"
    elif type_def.kind == "map":
        name = generate_map_type(type_def, types, spec, name_context)
    elif type_def.kind == "base":
        name = lsp_to_base_types(type_def)
    elif type_def.kind == "literal":
        name = generate_literal_type(type_def, types, spec, name_context)
    elif type_def.kind == "stringLiteral":
        name = "string"
    elif type_def.kind == "tuple":
        subset = filter_null_base_type(type_def.items)
        subset_types = [
            get_type_name(item, types, spec, name_context) for item in subset
        ]
        name = f"({', '.join(subset_types)})"
    elif type_def.kind == "or":
        subset = filter_null_base_type(type_def.items)
        if len(subset) == 1:
            name = get_type_name(subset[0], types, spec, name_context)
        elif len(subset) >= 2:
            if are_variant_literals(subset):
                name = generate_class_from_variant_literals(
                    subset, spec, types, name_context
                )
            else:
                subset_types = [
                    get_type_name(item, types, spec, name_context) for item in subset
                ]
                name = f"OrType<{', '.join(subset_types)}>"
        else:
            raise ValueError(f"Unknown type kind: {type_def.kind}")
    else:
        raise ValueError(f"Unknown type kind: {type_def.kind}")
    return name


def generate_map_type(
    type_def: model.LSP_TYPE_SPEC,
    types: TypeData,
    spec: model.LSPModel,
    name_context: Optional[str] = None,
) -> str:
    key_type = get_type_name(type_def.key, types, spec, name_context)

    if type_def.value.kind == "or":
        subset = filter_null_base_type(type_def.value.items)
        if len(subset) == 1:
            value_type = get_type_name(type_def.value, types, spec, name_context)
        else:
            value_type = to_upper_camel_case(f"{name_context}Value")
            type_alias = model.TypeAlias(
                **{
                    "name": value_type,
                    "type": type_def.value,
                }
            )
            generate_class_from_type_alias(type_alias, spec, types)

    else:
        value_type = get_type_name(type_def.value, types, spec, name_context)
    return f"ImmutableDictionary<{key_type}, {value_type}>"


def get_converter(type_def: model.LSP_TYPE_SPEC, type_name: str) -> Optional[str]:
    if type_def.kind == "base" and type_def.name in ["DocumentUri", "URI"]:
        return "[JsonConverter(typeof(CustomStringConverter<Uri>))]"
    elif type_def.kind == "reference" and type_def.name in [
        "Pattern",
        "ChangeAnnotationIdentifier",
    ]:
        return f"[JsonConverter(typeof(CustomStringConverter<{type_def.name}>))]"
    elif type_def.kind == "reference" and type_def.name == "DocumentSelector":
        return "[JsonConverter(typeof(DocumentSelectorConverter))]"
    elif type_def.kind == "or":
        subset = filter_null_base_type(type_def.items)
        if len(subset) == 1:
            return get_converter(subset[0], type_name)
        elif len(subset) >= 2:
            converter = type_name.replace("OrType<", "OrTypeConverter<")
            return f"[JsonConverter(typeof({converter}))]"
    elif type_def.kind == "array" and type_name.startswith("OrType<"):
        matches = ORTYPE_CONVERTER_RE.match(type_name).groupdict()
        if "parts" in matches:
            converter = f"OrTypeArrayConverter<{matches['parts']}>"
            return f"[JsonConverter(typeof({converter}))]"
    elif type_def.kind == "array":
        matches = IMMUTABLE_ARRAY_CONVERTER_RE.match(type_name).groupdict()
        elements = matches["elements"]
        if elements.startswith("OrType<"):
            matches = ORTYPE_CONVERTER_RE.match(elements).groupdict()
            converter = f"OrTypeArrayConverter<{matches['parts']}>"
            return f"[JsonConverter(typeof({converter}))]"
        else:
            converter = f"CustomArrayConverter<{elements}>"
            return f"[JsonConverter(typeof({converter}))]"
    return None


def generate_property(
    prop_def: model.Property,
    spec: model.LSPModel,
    types: TypeData,
    usings: List[str],
    class_name: str = "",
) -> Tuple[List[str], str]:
    if prop_def.name == "jsonrpc":
        name = "JsonRPC"
    else:
        name = to_upper_camel_case(prop_def.name)
    type_name = get_type_name(
        prop_def.type, types, spec, f"{class_name}_{prop_def.name}"
    )
    converter = get_converter(prop_def.type, type_name)
    special_optional = prop_def.type.kind == "or" and has_null_base_type(
        prop_def.type.items
    )
    optional = (
        "?"
        if (prop_def.optional or special_optional)
        and not (
            type_name.startswith("ImmutableArray<")
            or type_name.startswith("ImmutableDictionary<")
        )
        else ""
    )
    lines = (
        get_doc(prop_def.documentation)
        + generate_extras(prop_def)
        + ([converter] if converter else [])
        + (
            ["[JsonProperty(NullValueHandling = NullValueHandling.Ignore)]"]
            if optional and not special_optional
            else []
        )
        + [
            f'[DataMember(Name = "{prop_def.name}")]',
        ]
    )

    if prop_def.type.kind == "stringLiteral":
        lines.append(
            f'public {type_name}{optional} {name} {{ get; init; }} = "{prop_def.type.value}";'
        )
    else:
        lines.append(f"public {type_name}{optional} {name} {{ get; init; }}")

    usings.append("DataMember")
    if converter:
        usings.append("JsonConverter")
    if optional and not special_optional:
        usings.append("JsonProperty")

    return lines, type_name


def generate_name(name_context: str, types: TypeData) -> str:
    # If name context has a '_' it is likely a property.
    # Try name generation using just the property name
    parts = [to_upper_camel_case(p) for p in name_context.split("_") if len(p) > 3]

    # Try the last part of the name context
    name = parts[-1]
    if not types.get_by_name(name) and "info" in name_context.lower():
        return name

    # Combine all parts and try again
    name = "".join(parts)
    if not types.get_by_name(name):
        return name

    raise ValueError(f"Unable to generate name for {name_context}")


def generate_literal_type(
    literal: model.LiteralType,
    types: TypeData,
    spec: model.LSPModel,
    name_context: Optional[str] = None,
) -> str:
    if len(literal.value.properties) == 0:
        return "LSPObject"

    if types.get_by_name(literal.name) and not _get_struct(literal.name, spec):
        return literal.name

    if name_context is None:
        raise ValueError("name_context must be provided for literal types")

    if name_context.startswith("I") and name_context[1].isupper():
        # This is a interface name ISomething => Something
        name_context = name_context[1:]

    if "_" not in name_context:
        name_context = f"{name_context}_{get_context_from_literal(literal)}"

    literal.name = generate_name(name_context, types)

    usings = ["DataContract"]
    inner = []
    for prop in literal.value.properties:
        prop_code, _ = generate_property(prop, spec, types, usings, literal.name)
        inner += prop_code

    lines = namespace_wrapper(
        NAMESPACE,
        get_usings(usings + get_types_for_usings(inner)),
        class_wrapper(literal, inner),
    )
    types.add_type_info(literal, literal.name, lines)
    return literal.name


def generate_constructor(
    struct: model.Structure,
    types: TypeData,
    properties: List[Tuple[model.Property, str]],
) -> List[str]:
    class_name = get_special_case_class_name(struct.name)
    constructor = [
        "[JsonConstructor]",
        f"public {class_name}(",
    ]

    arguments = []
    optional_args = []
    assignments = []
    ctor_data = []
    for prop, prop_type in properties:
        name = get_special_case_property_name(to_camel_case(prop.name))
        special_optional = prop.type.kind == "or" and has_null_base_type(
            prop.type.items
        )
        if prop.optional or special_optional:
            if prop_type.startswith("ImmutableArray<") or prop_type.startswith(
                "ImmutableDictionary<"
            ):
                optional_args += [f"{prop_type} {name} = default!"]
            else:
                optional_args += [f"{prop_type}? {name} = null"]
            ctor_data += [(prop_type, name, True)]
        elif prop.name == "jsonrpc":
            optional_args += [f'{prop_type} {name} = "2.0"']
            ctor_data += [(prop_type, name, True)]
        else:
            arguments += [f"{prop_type} {name}"]
            ctor_data += [(prop_type, name, False)]

        if prop.name == "jsonrpc":
            assignments += [f"JsonRPC = {name};"]
        else:
            assignments += [f"{to_upper_camel_case(prop.name)} = {name};"]

    # combine args with a '\n' to get comma with indent
    all_args = (",\n".join(indent_lines(arguments + optional_args))).splitlines()
    types.add_ctor(struct.name, ctor_data)

    # re-split args to get the right coma placement and indent
    constructor += all_args
    constructor += [")", "{"]
    constructor += indent_lines(assignments)
    constructor += ["}"]
    return constructor


def generate_class_from_struct(
    struct: model.Structure,
    spec: model.LSPModel,
    types: TypeData,
    derived: Optional[str] = None,
    attributes: Optional[List[str]] = None,
):
    if types.get_by_name(struct.name) or struct.name.startswith("_"):
        return

    if attributes is None:
        attributes = []

    inner = []
    usings = ["DataContract", "JsonConstructor"]

    properties = get_all_properties(struct, spec)
    prop_types = []
    for prop in properties:
        prop_code, prop_type = generate_property(prop, spec, types, usings, struct.name)
        inner += prop_code
        prop_types += [prop_type]

    ctor = generate_constructor(struct, types, zip(properties, prop_types))
    inner = ctor + inner

    lines = namespace_wrapper(
        NAMESPACE,
        get_usings(usings + get_types_for_usings(inner + attributes)),
        class_wrapper(struct, inner, derived, attributes),
    )
    types.add_type_info(struct, struct.name, lines)


def get_context_from_literal(literal: model.LiteralType) -> str:
    if len(literal.value.properties) == 0:
        return "LSPObject"

    skipped = 0
    skip = [
        "range",
        "rangeLength",
        "position",
        "position",
        "location",
        "locationLink",
        "text",
    ]
    for prop in literal.value.properties:
        if prop.name in skip:
            skipped += 1
            continue
        return prop.name

    if skipped == len(literal.value.properties):
        # pick property with longest name
        names = sorted([p.name for p in literal.value.properties])
        return sorted(names, key=lambda n: len(n))[-1]

    return ""


def generate_type_alias_constructor(
    type_def: model.TypeAlias, spec: model.LSPModel, types: TypeData
) -> List[str]:
    constructor = []

    if type_def.type.kind == "or":
        subset = filter_null_base_type(type_def.type.items)
        if len(subset) == 1:
            raise ValueError("Unable to generate constructor for single item union")
        elif len(subset) >= 2:
            type_name = to_upper_camel_case(type_def.name)
            for t in subset:
                sub_type = get_type_name(t, types, spec, type_def.name)
                arg = get_special_case_property_name(to_camel_case(sub_type))
                matches = re.match(r"ImmutableArray<(?P<arg>\w+)>", arg)
                if matches:
                    arg = f"{matches['arg']}s"

                constructor += [
                    f"public {type_name}({sub_type} {arg}): base({arg}) {{}}",
                ]
        else:
            raise ValueError("Unable to generate constructor for empty union")
    elif type_def.type.kind == "reference":
        type_name = to_upper_camel_case(type_def.name)
        ctor_data = types.get_ctor(type_def.type.name)
        required = [
            (prop_type, prop_name)
            for prop_type, prop_name, optional in ctor_data
            if not optional
        ]
        optional = [
            (prop_type, prop_name)
            for prop_type, prop_name, optional in ctor_data
            if optional
        ]

        ctor_args = [f"{prop_type} {prop_name}" for prop_type, prop_name in required]
        ctor_args += [
            f"{prop_type}? {prop_name} = null" for prop_type, prop_name in optional
        ]

        base_args = [f"{prop_name}" for _, prop_name in required + optional]
        constructor += [
            f"public {type_name}({','.join(ctor_args)}): base({','.join(base_args)}) {{}}",
        ]

    return constructor


def generate_type_alias_converter(
    type_def: model.TypeAlias, spec: model.LSPModel, types: TypeData
) -> None:
    assert type_def.type.kind == "or"
    subset_types = [
        get_type_name(i, types, spec, type_def.name)
        for i in filter_null_base_type(type_def.type.items)
    ]
    converter = f"{type_def.name}Converter"
    or_type_converter = f"OrTypeConverter<{','.join(subset_types)}>"
    or_type = f"OrType<{','.join(subset_types)}>"
    code = [
        f"public class {converter} : JsonConverter<{type_def.name}>",
        "{",
        f"private {or_type_converter} _orType;",
        f"public {converter}()",
        "{",
        f"_orType = new {or_type_converter}();",
        "}",
        f"public override {type_def.name}? ReadJson(JsonReader reader, Type objectType, {type_def.name}? existingValue, bool hasExistingValue, JsonSerializer serializer)",
        "{",
        "reader = reader ?? throw new ArgumentNullException(nameof(reader));",
        "if (reader.TokenType == JsonToken.Null) { return null; }",
        f"var o = _orType.ReadJson(reader, objectType, existingValue, serializer);",
        f"if (o is {or_type} orType)",
        "{",
    ]
    for t in subset_types:
        code += [
            f"if (orType.Value?.GetType() == typeof({t}))",
            "{",
            f"return new {type_def.name}(({t})orType.Value);",
            "}",
        ]
    code += [
        "}",
        'throw new JsonSerializationException($"Unexpected token type.");',
        "}",
        f"public override void WriteJson(JsonWriter writer, {type_def.name}? value, JsonSerializer serializer)",
        "{",
        "_orType.WriteJson(writer, value, serializer);",
        "}",
        "}",
    ]

    code = namespace_wrapper(
        NAMESPACE, get_usings(["JsonConverter"] + get_types_for_usings(code)), code
    )

    ref = model.Structure(**{"name": converter, "properties": []})
    types.add_type_info(ref, converter, code)
    return converter


def generate_class_from_type_alias(
    type_def: model.TypeAlias, spec: model.LSPModel, types: TypeData
) -> None:
    if types.get_by_name(type_def.name):
        return

    usings = ["DataContract"]
    type_name = get_type_name(type_def.type, types, spec, type_def.name)
    class_attributes = []
    if type_def.type.kind == "or":
        converter = generate_type_alias_converter(type_def, spec, types)
        class_attributes += [f"[JsonConverter(typeof({converter}))]"]
        usings.append("JsonConverter")

    inner = generate_type_alias_constructor(type_def, spec, types)
    lines = namespace_wrapper(
        NAMESPACE,
        get_usings(usings + get_types_for_usings(inner)),
        class_wrapper(type_def, inner, type_name, class_attributes),
    )
    types.add_type_info(type_def, type_def.name, lines)


def generate_class_from_variant_literals(
    literals: List[model.LiteralType],
    spec: model.LSPModel,
    types: TypeData,
    name_context: Optional[str] = None,
) -> str:
    name = generate_name(name_context, types)
    if types.get_by_name(name):
        raise ValueError(f"Name {name} already exists")

    struct = model.Structure(
        **{
            "name": name,
            "properties": get_properties_from_literals(literals),
        }
    )

    lines = generate_code_for_variant_struct(struct, spec, types)
    types.add_type_info(struct, struct.name, lines)
    return struct.name


def get_properties_from_literals(literals: List[model.LiteralType]) -> Dict[str, Any]:
    properties = []
    for literal in literals:
        assert literal.kind == "literal"
        for prop in literal.value.properties:
            if prop.name not in [p["name"] for p in properties]:
                properties.append(
                    {
                        "name": prop.name,
                        "type": cattrs.unstructure(prop.type),
                        "optional": has_optional_variant(literals, prop.name),  #
                    }
                )
    return properties


def generate_code_for_variant_struct(
    struct: model.Structure,
    spec: model.LSPModel,
    types: TypeData,
) -> None:
    prop_types = []
    inner = []
    usings = ["DataContract", "JsonConstructor"]
    for prop in struct.properties:
        prop_code, prop_type = generate_property(prop, spec, types, usings, struct.name)
        inner += prop_code
        prop_types += [prop_type]

    ctor_data = []
    constructor_args = []
    conditions = []
    for prop, prop_type in zip(struct.properties, prop_types):
        name = get_special_case_property_name(to_camel_case(prop.name))
        immutable = prop_type.startswith("ImmutableArray<") or prop_type.startswith(
            "ImmutableDictionary<"
        )
        constructor_args += [
            f"{prop_type} {name}" if immutable else f"{prop_type}? {name}"
        ]
        ctor_data = [(prop_type)]
        if immutable:
            conditions += [f"({name}.IsDefault)"]
        else:
            conditions += [f"({name} is null)"]

    sig = ", ".join(constructor_args)
    types.add_ctor(struct.name, ctor_data)
    ctor = [
        f"[JsonConstructor]",
        f"public {struct.name}({sig})",
        "{",
        *indent_lines(
            [
                f"if ({'&&'.join(conditions)})",
                "{",
                *indent_lines(
                    [
                        f'throw new ArgumentException("At least one of the arguments must be non-null");'
                    ]
                ),
                "}",
            ]
        ),
        *indent_lines(
            [
                f"{to_upper_camel_case(prop.name)} = {get_special_case_property_name(to_camel_case(prop.name))};"
                for prop in struct.properties
            ]
        ),
        "}",
    ]

    inner = ctor + inner

    return namespace_wrapper(
        NAMESPACE,
        get_usings(usings + get_types_for_usings(inner)),
        class_wrapper(struct, inner, None),
    )


def generate_class_from_variant_type_alias(
    type_def: model.TypeAlias,
    spec: model.LSPModel,
    types: TypeData,
    name_context: Optional[str] = None,
) -> None:
    struct = model.Structure(
        **{
            "name": type_def.name,
            "properties": get_properties_from_literals(type_def.type.items),
            "documentation": type_def.documentation,
            "since": type_def.since,
            "deprecated": type_def.deprecated,
            "proposed": type_def.proposed,
        }
    )

    lines = generate_code_for_variant_struct(struct, spec, types)
    types.add_type_info(type_def, type_def.name, lines)


def has_optional_variant(literals: List[model.LiteralType], property_name: str) -> bool:
    count = 0
    optional = False
    for literal in literals:
        for prop in literal.value.properties:
            if prop.name == property_name:
                count += 1
                optional = optional or prop.optional
    return optional and count == len(literals)


def are_variant_literals(literals: List[model.LiteralType]) -> bool:
    if all(i.kind == "literal" for i in literals):
        return all(
            has_optional_variant(literals, prop.name)
            for prop in literals[0].value.properties
        )
    return False


def is_variant_type_alias(type_def: model.TypeAlias) -> bool:
    if type_def.type.kind == "or" and all(
        i.kind == "literal" for i in type_def.type.items
    ):
        literals = type_def.type.items
        return all(
            has_optional_variant(literals, prop.name)
            for prop in literals[0].value.properties
        )
    return False


def copy_struct(struct_def: model.Structure, new_name: str):
    converter = cattrs.GenConverter()
    obj = converter.unstructure(struct_def, model.Structure)
    obj["name"] = new_name
    return model.Structure(**obj)


def copy_property(prop_def: model.Property):
    converter = cattrs.GenConverter()
    obj = converter.unstructure(prop_def, model.Property)
    return model.Property(**obj)


def get_all_extends(struct_def: model.Structure, spec) -> List[model.Structure]:
    extends = []
    for extend in struct_def.extends:
        extends.append(_get_struct(extend.name, spec))
        for struct in get_all_extends(_get_struct(extend.name, spec), spec):
            if not any(struct.name == e.name for e in extends):
                extends.append(struct)
    return extends


def get_all_properties(struct: model.Structure, spec) -> List[model.Structure]:
    properties = []
    for prop in struct.properties:
        properties.append(copy_property(prop))

    for extend in get_all_extends(struct, spec):
        for prop in get_all_properties(extend, spec):
            if not any(prop.name == p.name for p in properties):
                properties.append(copy_property(prop))

    if not all(mixin.kind == "reference" for mixin in struct.mixins):
        raise ValueError(f"Struct {struct.name} has non-reference mixins")
    for mixin in [_get_struct(mixin.name, spec) for mixin in struct.mixins]:
        for prop in get_all_properties(mixin, spec):
            if not any(prop.name == p.name for p in properties):
                properties.append(copy_property(prop))

    return properties


def generate_code_for_request(request: model.Request):
    lines = get_doc(request.documentation) + generate_extras(request)
    lines.append(
        f'public static string {lsp_method_to_name(request.method)} {{ get; }} = "{request.method}";'
    )
    return lines


def generate_code_for_notification(notify: model.Notification):
    lines = get_doc(notify.documentation) + generate_extras(notify)
    lines.append(
        f'public static string {lsp_method_to_name(notify.method)} {{ get; }} = "{notify.method}";'
    )
    return lines


def generate_request_notification_methods(spec: model.LSPModel, types: TypeData):
    inner_lines = []
    for request in spec.requests:
        inner_lines += generate_code_for_request(request)

    for notification in spec.notifications:
        inner_lines += generate_code_for_notification(notification)

    lines = namespace_wrapper(
        NAMESPACE,
        get_usings(["System"] + get_types_for_usings(inner_lines)),
        ["public static class LSPMethods", "{", *indent_lines(inner_lines), "}"],
    )
    enum_type = model.Enum(
        **{
            "name": "LSPMethods",
            "type": {"kind": "base", "name": "string"},
            "values": [],
            "documentation": "LSP methods as defined in the LSP spec",
        }
    )
    types.add_type_info(enum_type, "LSPMethods", lines)


def get_message_template(
    obj: Union[model.Request, model.Notification],
    is_request: bool,
) -> model.Structure:
    text = "Request" if is_request else "Notification"
    properties = [
        {
            "name": "jsonrpc",
            "type": {"kind": "stringLiteral", "value": "2.0"},
            "documentation": "The jsonrpc version.",
        }
    ]
    if is_request:
        properties += [
            {
                "name": "id",
                "type": {
                    "kind": "or",
                    "items": [
                        {"kind": "base", "name": "string"},
                        {"kind": "base", "name": "integer"},
                    ],
                },
                "documentation": f"The {text} id.",
            }
        ]
    properties += [
        {
            "name": "method",
            "type": {"kind": "base", "name": "string"},
            "documentation": f"The {text} method.",
        },
    ]
    if obj.params:
        properties.append(
            {
                "name": "params",
                "type": cattrs.unstructure(obj.params),
                "documentation": f"The {text} parameters.",
            }
        )
    else:
        properties.append(
            {
                "name": "params",
                "type": {"kind": "reference", "name": "LSPAny"},
                "documentation": f"The {text} parameters.",
                "optional": True,
            }
        )

    class_template = {
        "name": f"{lsp_method_to_name(obj.method)}{text}",
        "properties": properties,
        "documentation": obj.documentation,
        "since": obj.since,
        "deprecated": obj.deprecated,
        "proposed": obj.proposed,
    }
    return model.Structure(**class_template)


def get_response_template(
    obj: model.Request, spec: model.LSPModel, types: TypeData
) -> model.Structure:
    properties = [
        {
            "name": "jsonrpc",
            "type": {"kind": "stringLiteral", "value": "2.0"},
            "documentation": "The jsonrpc version.",
        },
        {
            "name": "id",
            "type": {
                "kind": "or",
                "items": [
                    {"kind": "base", "name": "string"},
                    {"kind": "base", "name": "integer"},
                ],
            },
            "documentation": f"The Request id.",
        },
    ]
    if obj.result:
        properties.append(
            {
                "name": "result",
                "type": cattrs.unstructure(obj.result),
                "documentation": f"Results for the request.",
                "optional": True,
            }
        )
    else:
        properties.append(
            {
                "name": "result",
                "type": {"kind": "base", "name": "null"},
                "documentation": f"Results for the request.",
                "optional": True,
            }
        )
    properties.append(
        {
            "name": "error",
            "type": {"kind": "reference", "name": "ResponseError"},
            "documentation": f"Error while handling the request.",
            "optional": True,
        }
    )
    class_template = {
        "name": f"{lsp_method_to_name(obj.method)}Response",
        "properties": properties,
        "documentation": obj.documentation,
        "since": obj.since,
        "deprecated": obj.deprecated,
        "proposed": obj.proposed,
    }
    return model.Structure(**class_template)


def get_registration_options_template(
    obj: Union[model.Request, model.Notification],
    spec: model.LSPModel,
    types: TypeData,
) -> model.Structure:
    if obj.registrationOptions and obj.registrationOptions.kind != "reference":
        if obj.registrationOptions.kind == "and":
            structs = [_get_struct(s.name, spec) for s in obj.registrationOptions.items]
            properties = []
            for struct in structs:
                properties += get_all_properties(struct, spec)

            class_template = {
                "name": f"{lsp_method_to_name(obj.method)}RegistrationOptions",
                "properties": [
                    cattrs.unstructure(p, model.Property) for p in properties
                ],
            }
            return model.Structure(**class_template)
        else:
            raise ValueError(
                f"Unexpected registrationOptions type: {obj.registrationOptions.type.kind}"
            )
    return None


def generate_all_classes(spec: model.LSPModel, types: TypeData):
    for struct in spec.structures:
        generate_class_from_struct(struct, spec, types)

    for type_alias in spec.typeAliases:
        if is_variant_type_alias(type_alias):
            generate_class_from_variant_type_alias(type_alias, spec, types)
        else:
            generate_class_from_type_alias(type_alias, spec, types)

    generate_request_notification_methods(spec, types)

    for request in spec.requests:
        partial_result_name = None
        if request.partialResult:
            partial_result_name = get_type_name(request.partialResult, types, spec)

        struct = get_message_template(request, is_request=True)
        generate_class_from_struct(
            struct,
            spec,
            types,
            (
                f"IRequest<{get_type_name(request.params, types, spec)}>"
                if request.params
                else "IRequest<LSPAny?>"
            ),
            [
                f"[Direction(MessageDirection.{to_upper_camel_case(request.messageDirection)})]",
                f'[LSPRequest("{request.method}", typeof({lsp_method_to_name(request.method)}Response), typeof({partial_result_name}))]'
                if partial_result_name
                else f'[LSPRequest("{request.method}", typeof({lsp_method_to_name(request.method)}Response))]',
            ],
        )
        response = get_response_template(request, spec, types)
        generate_class_from_struct(
            response,
            spec,
            types,
            f"IResponse<{get_type_name(request.result, types, spec)}>",
            [
                f"[LSPResponse(typeof({lsp_method_to_name(request.method)}Request))]",
            ],
        )
        registration_options = get_registration_options_template(request, spec, types)
        if registration_options:
            generate_class_from_struct(
                registration_options,
                spec,
                types,
            )

    for notification in spec.notifications:
        struct = get_message_template(notification, is_request=False)
        generate_class_from_struct(
            struct,
            spec,
            types,
            (
                f"INotification<{get_type_name(notification.params, types, spec)}>"
                if notification.params
                else "INotification<LSPAny>"
            ),
            [
                f"[Direction(MessageDirection.{to_upper_camel_case(request.messageDirection)})]",
            ],
        )
        registration_options = get_registration_options_template(
            notification, spec, types
        )
        if registration_options:
            generate_class_from_struct(
                registration_options,
                spec,
                types,
            )



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import logging
import os
import pathlib
import subprocess
from typing import Dict, List

import generator.model as model

from .dotnet_classes import generate_all_classes
from .dotnet_commons import TypeData
from .dotnet_constants import NAMESPACE, PACKAGE_DIR_NAME
from .dotnet_enums import generate_enums
from .dotnet_helpers import namespace_wrapper
from .dotnet_special_classes import generate_special_classes

LOGGER = logging.getLogger("dotnet")


def generate_from_spec(spec: model.LSPModel, output_dir: str) -> None:
    """Generate the code for the given spec."""
    output_path = pathlib.Path(output_dir, PACKAGE_DIR_NAME)
    if not output_path.exists():
        output_path.mkdir(parents=True, exist_ok=True)

    cleanup(output_path)
    copy_custom_classes(output_path)

    LOGGER.info("Generating code in C#")
    types = TypeData()
    generate_package_code(spec, types)

    for name, lines in types.get_all():
        file_name = f"{name}.cs"
        (output_path / file_name).write_text("\n".join(lines), encoding="utf-8")


def generate_package_code(spec: model.LSPModel, types: TypeData) -> Dict[str, str]:
    generate_enums(spec, types)
    generate_special_classes(spec, types)
    generate_all_classes(spec, types)


def cleanup(output_path: pathlib.Path) -> None:
    """Cleanup the generated C# files."""
    for file in output_path.glob("*.cs"):
        file.unlink()


def copy_custom_classes(output_path: pathlib.Path) -> None:
    """Copy the custom classes to the output directory."""
    custom = pathlib.Path(__file__).parent / "custom"
    for file in custom.glob("*.cs"):
        lines = file.read_text(encoding="utf-8").splitlines()
        lines = namespace_wrapper(NAMESPACE, [], lines)
        (output_path / file.name).write_text("\n".join(lines), encoding="utf-8")



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import logging
import pathlib
from typing import Dict, List

import generator.model as model

from .testdata_generator import generate

logger = logging.getLogger("testdata")


def generate_from_spec(spec: model.LSPModel, output_dir: str) -> None:
    """Generate the code for the given spec."""
    output = pathlib.Path(output_dir)

    if not output.exists():
        output.mkdir(parents=True, exist_ok=True)

    cleanup(output)
    # key is the relative path to the file, value is the content
    code: Dict[str, str] = generate(spec, logger)
    for file_name in code:
        # print file size
        file = output / file_name
        file.write_text(code[file_name], encoding="utf-8")


def cleanup(output_path: pathlib.Path) -> None:
    """Cleanup the generated C# files."""
    for file in output_path.glob("*.json"):
        file.unlink()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json
import uuid

import hamcrest
import jsonrpc
import pytest

ID = str(uuid.uuid4())
INITIALIZE_PARAMS = {
    "processId": 1105947,
    "rootPath": "/home/user/src/Personal/jedi-language-server",
    "rootUri": "file:///home/user/src/Personal/jedi-language-server",
    "capabilities": {
        "workspace": {
            "applyEdit": True,
            "workspaceEdit": {
                "documentChanges": True,
                "resourceOperations": ["create", "rename", "delete"],
                "failureHandling": "undo",
                "normalizesLineEndings": True,
                "changeAnnotationSupport": {"groupsOnLabel": False},
            },
            "didChangeConfiguration": {"dynamicRegistration": True},
            "didChangeWatchedFiles": {
                "dynamicRegistration": True,
                "relativePatternSupport": True,
            },
            "codeLens": {"refreshSupport": True},
            "executeCommand": {"dynamicRegistration": True},
            "configuration": True,
            "fileOperations": {
                "dynamicRegistration": True,
                "didCreate": True,
                "didRename": True,
                "didDelete": True,
                "willCreate": True,
                "willRename": True,
                "willDelete": True,
            },
            "semanticTokens": {"refreshSupport": True},
            "inlayHint": {"refreshSupport": True},
            "inlineValue": {"refreshSupport": True},
            "diagnostics": {"refreshSupport": True},
            "symbol": {
                "dynamicRegistration": True,
                "symbolKind": {
                    "valueSet": [
                        1,
                        2,
                        3,
                        4,
                        5,
                        6,
                        7,
                        8,
                        9,
                        10,
                        11,
                        12,
                        13,
                        14,
                        15,
                        16,
                        17,
                        18,
                        19,
                        20,
                        21,
                        22,
                        23,
                        24,
                        25,
                        26,
                    ]
                },
                "tagSupport": {"valueSet": [1]},
                "resolveSupport": {"properties": ["location.range"]},
            },
            "workspaceFolders": True,
        },
        "textDocument": {
            "publishDiagnostics": {
                "relatedInformation": True,
                "versionSupport": True,
                "tagSupport": {"valueSet": [1, 2]},
                "codeDescriptionSupport": True,
                "dataSupport": True,
            },
            "synchronization": {
                "dynamicRegistration": True,
                "willSave": True,
                "willSaveWaitUntil": True,
                "didSave": True,
            },
            "completion": {
                "dynamicRegistration": True,
                "contextSupport": True,
                "completionItem": {
                    "snippetSupport": True,
                    "commitCharactersSupport": True,
                    "documentationFormat": ["markdown", "plaintext"],
                    "deprecatedSupport": True,
                    "preselectSupport": True,
                    "insertReplaceSupport": True,
                    "tagSupport": {"valueSet": [1]},
                    "resolveSupport": {
                        "properties": ["documentation", "detail", "additionalTextEdits"]
                    },
                    "labelDetailsSupport": True,
                    "insertTextModeSupport": {"valueSet": [1, 2]},
                },
                "completionItemKind": {
                    "valueSet": [
                        1,
                        2,
                        3,
                        4,
                        5,
                        6,
                        7,
                        8,
                        9,
                        10,
                        11,
                        12,
                        13,
                        14,
                        15,
                        16,
                        17,
                        18,
                        19,
                        20,
                        21,
                        22,
                        23,
                        24,
                        25,
                    ]
                },
                "insertTextMode": 2,
                "completionList": {
                    "itemDefaults": [
                        "commitCharacters",
                        "editRange",
                        "insertTextFormat",
                        "insertTextMode",
                    ]
                },
            },
            "hover": {
                "dynamicRegistration": True,
                "contentFormat": ["markdown", "plaintext"],
            },
            "signatureHelp": {
                "dynamicRegistration": True,
                "contextSupport": True,
                "signatureInformation": {
                    "documentationFormat": ["markdown", "plaintext"],
                    "activeParameterSupport": True,
                    "parameterInformation": {"labelOffsetSupport": True},
                },
            },
            "references": {"dynamicRegistration": True},
            "definition": {"dynamicRegistration": True, "linkSupport": True},
            "documentHighlight": {"dynamicRegistration": True},
            "documentSymbol": {
                "dynamicRegistration": True,
                "symbolKind": {
                    "valueSet": [
                        1,
                        2,
                        3,
                        4,
                        5,
                        6,
                        7,
                        8,
                        9,
                        10,
                        11,
                        12,
                        13,
                        14,
                        15,
                        16,
                        17,
                        18,
                        19,
                        20,
                        21,
                        22,
                        23,
                        24,
                        25,
                        26,
                    ]
                },
                "hierarchicalDocumentSymbolSupport": True,
                "tagSupport": {"valueSet": [1]},
                "labelSupport": True,
            },
            "codeAction": {
                "dynamicRegistration": True,
                "isPreferredSupport": True,
                "disabledSupport": True,
                "dataSupport": True,
                "honorsChangeAnnotations": False,
                "resolveSupport": {"properties": ["edit"]},
                "codeActionLiteralSupport": {
                    "codeActionKind": {
                        "valueSet": [
                            "",
                            "quickfix",
                            "refactor",
                            "refactor.extract",
                            "refactor.inline",
                            "refactor.rewrite",
                            "source",
                            "source.organizeImports",
                        ]
                    }
                },
            },
            "codeLens": {"dynamicRegistration": True},
            "formatting": {"dynamicRegistration": True},
            "rangeFormatting": {"dynamicRegistration": True},
            "onTypeFormatting": {"dynamicRegistration": True},
            "rename": {
                "dynamicRegistration": True,
                "prepareSupport": True,
                "honorsChangeAnnotations": True,
                "prepareSupportDefaultBehavior": 1,
            },
            "documentLink": {"dynamicRegistration": True, "tooltipSupport": True},
            "typeDefinition": {"dynamicRegistration": True, "linkSupport": True},
            "implementation": {"dynamicRegistration": True, "linkSupport": True},
            "declaration": {"dynamicRegistration": True, "linkSupport": True},
            "colorProvider": {"dynamicRegistration": True},
            "foldingRange": {
                "dynamicRegistration": True,
                "rangeLimit": 5000,
                "lineFoldingOnly": True,
                "foldingRangeKind": {"valueSet": ["comment", "imports", "region"]},
                "foldingRange": {"collapsedText": False},
            },
            "selectionRange": {"dynamicRegistration": True},
            "callHierarchy": {"dynamicRegistration": True},
            "linkedEditingRange": {"dynamicRegistration": True},
            "semanticTokens": {
                "dynamicRegistration": True,
                "tokenTypes": [
                    "namespace",
                    "type",
                    "class",
                    "enum",
                    "interface",
                    "struct",
                    "typeParameter",
                    "parameter",
                    "variable",
                    "property",
                    "enumMember",
                    "event",
                    "function",
                    "method",
                    "macro",
                    "keyword",
                    "modifier",
                    "comment",
                    "string",
                    "number",
                    "regexp",
                    "decorator",
                    "operator",
                ],
                "tokenModifiers": [
                    "declaration",
                    "definition",
                    "readonly",
                    "static",
                    "deprecated",
                    "abstract",
                    "async",
                    "modification",
                    "documentation",
                    "defaultLibrary",
                ],
                "formats": ["relative"],
                "requests": {"range": True, "full": {"delta": True}},
                "multilineTokenSupport": False,
                "overlappingTokenSupport": False,
                "serverCancelSupport": True,
                "augmentsSyntaxTokens": True,
            },
            "inlayHint": {
                "dynamicRegistration": True,
                "resolveSupport": {
                    "properties": [
                        "tooltip",
                        "textEdits",
                        "label.tooltip",
                        "label.location",
                        "label.command",
                    ]
                },
            },
            "inlineValue": {"dynamicRegistration": True},
            "diagnostic": {"dynamicRegistration": True, "relatedDocumentSupport": True},
            "typeHierarchy": {"dynamicRegistration": True},
        },
        "window": {
            "showMessage": {"messageActionItem": {"additionalPropertiesSupport": True}},
            "showDocument": {"support": True},
            "workDoneProgress": True,
        },
        "general": {
            "regularExpressions": {"engine": "ECMAScript", "version": "ES2020"},
            "markdown": {"parser": "marked", "version": "4.0.10"},
            "positionEncodings": ["utf-16"],
            "staleRequestSupport": {
                "cancel": True,
                "retryOnContentModified": [
                    "textDocument/inlayHint",
                    "textDocument/semanticTokens/full",
                    "textDocument/semanticTokens/range",
                    "textDocument/semanticTokens/full/delta",
                ],
            },
        },
    },
    "initializationOptions": {
        "enable": True,
        "startupMessage": False,
        "trace": {"server": "verbose"},
        "jediSettings": {
            "autoImportModules": ["pygls"],
            "caseInsensitiveCompletion": True,
            "debug": False,
        },
        "executable": {"args": [], "command": "jedi-language-server"},
        "codeAction": {
            "nameExtractFunction": "jls_extract_def",
            "nameExtractVariable": "jls_extract_var",
        },
        "completion": {
            "disableSnippets": False,
            "resolveEagerly": False,
            "ignorePatterns": [],
        },
        "diagnostics": {
            "enable": True,
            "didOpen": True,
            "didChange": True,
            "didSave": True,
        },
        "hover": {
            "enable": True,
            "disable": {
                "class": {"all": False, "names": [], "fullNames": []},
                "function": {"all": False, "names": [], "fullNames": []},
                "instance": {"all": False, "names": [], "fullNames": []},
                "keyword": {"all": False, "names": [], "fullNames": []},
                "module": {"all": False, "names": [], "fullNames": []},
                "param": {"all": False, "names": [], "fullNames": []},
                "path": {"all": False, "names": [], "fullNames": []},
                "property": {"all": False, "names": [], "fullNames": []},
                "statement": {"all": False, "names": [], "fullNames": []},
            },
        },
        "workspace": {
            "extraPaths": [],
            "symbols": {
                "maxSymbols": 20,
                "ignoreFolders": [".nox", ".tox", ".venv", "__pycache__", "venv"],
            },
        },
    },
    "trace": "verbose",
    "workspaceFolders": [
        {
            "uri": "file:///home/user/src/Personal/jedi-language-server",
            "name": "jedi-language-server",
        }
    ],
    "locale": "en_US",
    "clientInfo": {"name": "coc.nvim", "version": "0.0.82"},
}


TEST_DATA = [
    {"id": ID, "params": INITIALIZE_PARAMS, "method": "initialize", "jsonrpc": "2.0"},
]


@pytest.mark.parametrize("index", list(range(0, len(TEST_DATA))))
def test_initialize_request_params(index):
    data = TEST_DATA[index]
    data_str = json.dumps(data)
    parsed = jsonrpc.from_json(data_str)
    actual_str = jsonrpc.to_json(parsed)
    actual_data = json.loads(actual_str)
    hamcrest.assert_that(actual_data, hamcrest.is_(data))
